# --------------------------------------------------------------------

cdef extern from * nogil:

    ctypedef const char* PetscDMPlexTransformType "DMPlexTransformType"
    PetscDMPlexTransformType DMPLEXREFINEREGULAR
    PetscDMPlexTransformType DMPLEXREFINEALFELD
    PetscDMPlexTransformType DMPLEXREFINEPOWELLSABIN
    PetscDMPlexTransformType DMPLEXREFINEBOUNDARYLAYER
    PetscDMPlexTransformType DMPLEXREFINESBR
    PetscDMPlexTransformType DMPLEXREFINETOBOX
    PetscDMPlexTransformType DMPLEXREFINETOSIMPLEX
    PetscDMPlexTransformType DMPLEXREFINE1D
    PetscDMPlexTransformType DMPLEXEXTRUDETYPE
    PetscDMPlexTransformType DMPLEXTRANSFORMFILTER

    PetscErrorCode DMPlexCreate(MPI_Comm, PetscDM*)
    PetscErrorCode DMPlexCreateCohesiveSubmesh(PetscDM, PetscBool, const char[], PetscInt, PetscDM*)
    PetscErrorCode DMPlexCreateFromCellListPetsc(MPI_Comm, PetscInt, PetscInt, PetscInt, PetscInt, PetscBool, PetscInt[], PetscInt, PetscReal[], PetscDM*)

    PetscErrorCode DMPlexGetChart(PetscDM, PetscInt*, PetscInt*)
    PetscErrorCode DMPlexSetChart(PetscDM, PetscInt, PetscInt)
    PetscErrorCode DMPlexGetConeSize(PetscDM, PetscInt, PetscInt*)
    PetscErrorCode DMPlexSetConeSize(PetscDM, PetscInt, PetscInt)
    PetscErrorCode DMPlexGetCone(PetscDM, PetscInt, const PetscInt*[])
    PetscErrorCode DMPlexSetCone(PetscDM, PetscInt, const PetscInt[])
    PetscErrorCode DMPlexInsertCone(PetscDM, PetscInt, PetscInt, PetscInt)
    PetscErrorCode DMPlexInsertConeOrientation(PetscDM, PetscInt, PetscInt, PetscInt)
    PetscErrorCode DMPlexGetConeOrientation(PetscDM, PetscInt, const PetscInt*[])
    PetscErrorCode DMPlexSetConeOrientation(PetscDM, PetscInt, const PetscInt[])
    PetscErrorCode DMPlexSetCellType(PetscDM, PetscInt, PetscDMPolytopeType)
    PetscErrorCode DMPlexGetCellType(PetscDM, PetscInt, PetscDMPolytopeType*)
    PetscErrorCode DMPlexGetCellTypeLabel(PetscDM, PetscDMLabel*)
    PetscErrorCode DMPlexGetSupportSize(PetscDM, PetscInt, PetscInt*)
    PetscErrorCode DMPlexSetSupportSize(PetscDM, PetscInt, PetscInt)
    PetscErrorCode DMPlexGetSupport(PetscDM, PetscInt, const PetscInt*[])
    PetscErrorCode DMPlexSetSupport(PetscDM, PetscInt, const PetscInt[])
    PetscErrorCode DMPlexGetMaxSizes(PetscDM, PetscInt*, PetscInt*)
    PetscErrorCode DMPlexSymmetrize(PetscDM)
    PetscErrorCode DMPlexStratify(PetscDM)
    PetscErrorCode DMPlexOrient(PetscDM)
    PetscErrorCode DMPlexInterpolate(PetscDM, PetscDM*)
    PetscErrorCode DMPlexUninterpolate(PetscDM, PetscDM*)
    PetscErrorCode DMPlexGetPointLocal(PetscDM, PetscInt, PetscInt*, PetscInt*)
    PetscErrorCode DMPlexGetPointGlobal(PetscDM, PetscInt, PetscInt*, PetscInt*)
    PetscErrorCode DMPlexGetPointLocalField(PetscDM, PetscInt, PetscInt, PetscInt*, PetscInt*)
    PetscErrorCode DMPlexGetPointGlobalField(PetscDM, PetscInt, PetscInt, PetscInt*, PetscInt*)
    PetscErrorCode DMPlexCreateClosureIndex(PetscDM, PetscSection)

    PetscErrorCode DMPlexGetCellNumbering(PetscDM, PetscIS*)
    PetscErrorCode DMPlexGetVertexNumbering(PetscDM, PetscIS*)
    PetscErrorCode DMPlexCreatePointNumbering(PetscDM, PetscIS*)

    PetscErrorCode DMPlexGetDepth(PetscDM, PetscInt*)
    PetscErrorCode DMPlexGetDepthStratum(PetscDM, PetscInt, PetscInt*, PetscInt*)
    PetscErrorCode DMPlexGetHeightStratum(PetscDM, PetscInt, PetscInt*, PetscInt*)
    PetscErrorCode DMPlexGetPointDepth(PetscDM, PetscInt, PetscInt*)
    PetscErrorCode DMPlexGetPointHeight(PetscDM, PetscInt, PetscInt*)

    PetscErrorCode DMPlexGetMeet(PetscDM, PetscInt, const PetscInt[], PetscInt*, const PetscInt**)
    PetscErrorCode DMPlexRestoreMeet(PetscDM, PetscInt, const PetscInt[], PetscInt*, const PetscInt**)
    PetscErrorCode DMPlexGetJoin(PetscDM, PetscInt, const PetscInt[], PetscInt*, const PetscInt**)
    PetscErrorCode DMPlexGetFullJoin(PetscDM, PetscInt, const PetscInt[], PetscInt*, const PetscInt**)
    PetscErrorCode DMPlexRestoreJoin(PetscDM, PetscInt, const PetscInt[], PetscInt*, const PetscInt**)
    PetscErrorCode DMPlexGetTransitiveClosure(PetscDM, PetscInt, PetscBool, PetscInt*, PetscInt*[])
    PetscErrorCode DMPlexRestoreTransitiveClosure(PetscDM, PetscInt, PetscBool, PetscInt*, PetscInt*[])
    PetscErrorCode DMPlexVecGetClosure(PetscDM, PetscSection, PetscVec, PetscInt, PetscInt*, PetscScalar*[])
    PetscErrorCode DMPlexVecRestoreClosure(PetscDM, PetscSection, PetscVec, PetscInt, PetscInt*, PetscScalar*[])
    PetscErrorCode DMPlexVecSetClosure(PetscDM, PetscSection, PetscVec, PetscInt, PetscScalar[], PetscInsertMode)
    PetscErrorCode DMPlexMatSetClosure(PetscDM, PetscSection, PetscSection, PetscMat, PetscInt, PetscScalar[], PetscInsertMode)

    PetscErrorCode DMPlexGenerate(PetscDM, const char[], PetscBool , PetscDM*)
    PetscErrorCode DMPlexTriangleSetOptions(PetscDM, const char*)
    PetscErrorCode DMPlexTetgenSetOptions(PetscDM, const char*)
    PetscErrorCode DMPlexCreateBoxMesh(MPI_Comm, PetscInt, PetscBool, PetscInt[], PetscReal[], PetscReal[], PetscDMBoundaryType[], PetscBool, PetscInt, PetscBool, PetscDM*)
    PetscErrorCode DMPlexCreateBoxSurfaceMesh(MPI_Comm, PetscInt, PetscInt[], PetscReal[], PetscReal[], PetscBool, PetscDM*)
    PetscErrorCode DMPlexCreateFromFile(MPI_Comm, const char[], const char[], PetscBool, PetscDM*)
    PetscErrorCode DMPlexCreateCGNS(MPI_Comm, PetscInt, PetscBool, PetscDM*)
    PetscErrorCode DMPlexCreateCGNSFromFile(MPI_Comm, const char[], PetscBool, PetscDM*)
    PetscErrorCode DMPlexCreateExodus(MPI_Comm, int, PetscBool, PetscDM*)
    PetscErrorCode DMPlexCreateExodusFromFile(MPI_Comm, const char[], PetscBool, PetscDM*)
    PetscErrorCode DMPlexCreateGmsh(MPI_Comm, PetscViewer, PetscBool, PetscDM*)
    PetscErrorCode DMPlexCreateCoordinateSpace(PetscDM, PetscInt, PetscBool, PetscBool)

    PetscErrorCode DMPlexSetAdjacencyUseAnchors(PetscDM, PetscBool)
    PetscErrorCode DMPlexGetAdjacencyUseAnchors(PetscDM, PetscBool*)
    PetscErrorCode DMPlexGetAdjacency(PetscDM, PetscInt, PetscInt*, PetscInt*[])
    PetscErrorCode DMPlexRebalanceSharedPoints(PetscDM, PetscInt, PetscBool, PetscBool, PetscBool*)
    PetscErrorCode DMPlexDistribute(PetscDM, PetscInt, PetscSF*, PetscDM*)
    PetscErrorCode DMPlexDistributeOverlap(PetscDM, PetscInt, PetscSF*, PetscDM*)
    PetscErrorCode DMPlexDistributeGetDefault(PetscDM, PetscBool*)
    PetscErrorCode DMPlexDistributeSetDefault(PetscDM, PetscBool)
    PetscErrorCode DMPlexSetPartitioner(PetscDM, PetscPartitioner)
    PetscErrorCode DMPlexGetPartitioner(PetscDM, PetscPartitioner*)
    PetscErrorCode DMPlexDistributeField(PetscDM, PetscSF, PetscSection, PetscVec, PetscSection, PetscVec)
    PetscErrorCode DMPlexIsDistributed(PetscDM, PetscBool*)
    PetscErrorCode DMPlexIsSimplex(PetscDM, PetscBool*)
    PetscErrorCode DMPlexDistributionSetName(PetscDM, const char[])
    PetscErrorCode DMPlexDistributionGetName(PetscDM, const char*[])

    PetscErrorCode DMPlexGetOrdering(PetscDM, PetscMatOrderingType, PetscDMLabel, PetscIS*)
    PetscErrorCode DMPlexPermute(PetscDM, PetscIS, PetscDM*)
    PetscErrorCode DMPlexReorderGetDefault(PetscDM, PetscDMReorderDefaultFlag*)
    PetscErrorCode DMPlexReorderSetDefault(PetscDM, PetscDMReorderDefaultFlag)

    PetscErrorCode DMPlexGetSubpointMap(PetscDM, PetscDMLabel*)
    PetscErrorCode DMPlexGetSubpointIS(PetscDM, PetscIS*)

    PetscErrorCode DMPlexCreateCoarsePointIS(PetscDM, PetscIS*)
    PetscErrorCode DMPlexMarkBoundaryFaces(PetscDM, PetscInt, PetscDMLabel)
    PetscErrorCode DMPlexLabelComplete(PetscDM, PetscDMLabel)
    PetscErrorCode DMPlexLabelCohesiveComplete(PetscDM, PetscDMLabel, PetscDMLabel, PetscInt, PetscBool, PetscBool, PetscDM)

    PetscErrorCode DMPlexGetRefinementLimit(PetscDM, PetscReal*)
    PetscErrorCode DMPlexSetRefinementLimit(PetscDM, PetscReal)
    PetscErrorCode DMPlexGetRefinementUniform(PetscDM, PetscBool*)
    PetscErrorCode DMPlexSetRefinementUniform(PetscDM, PetscBool)

    PetscErrorCode DMPlexGetMinRadius(PetscDM, PetscReal*)

    PetscErrorCode DMPlexCreateSection(PetscDM, PetscDMLabel[], const PetscInt[], const PetscInt[], PetscInt, const PetscInt[], const PetscIS[], const PetscIS[], PetscIS, PetscSection*)

    PetscErrorCode DMPlexComputeCellGeometryFVM(PetscDM, PetscInt, PetscReal*, PetscReal[], PetscReal[])
    PetscErrorCode DMPlexConstructGhostCells(PetscDM, const char[], PetscInt*, PetscDM*)

    PetscErrorCode DMPlexMetricSetFromOptions(PetscDM)
    PetscErrorCode DMPlexMetricSetUniform(PetscDM, PetscBool)
    PetscErrorCode DMPlexMetricIsUniform(PetscDM, PetscBool*)
    PetscErrorCode DMPlexMetricSetIsotropic(PetscDM, PetscBool)
    PetscErrorCode DMPlexMetricIsIsotropic(PetscDM, PetscBool*)
    PetscErrorCode DMPlexMetricSetRestrictAnisotropyFirst(PetscDM, PetscBool)
    PetscErrorCode DMPlexMetricRestrictAnisotropyFirst(PetscDM, PetscBool*)
    PetscErrorCode DMPlexMetricSetNoInsertion(PetscDM, PetscBool)
    PetscErrorCode DMPlexMetricNoInsertion(PetscDM, PetscBool*)
    PetscErrorCode DMPlexMetricSetNoSwapping(PetscDM, PetscBool)
    PetscErrorCode DMPlexMetricNoSwapping(PetscDM, PetscBool*)
    PetscErrorCode DMPlexMetricSetNoMovement(PetscDM, PetscBool)
    PetscErrorCode DMPlexMetricNoMovement(PetscDM, PetscBool*)
    PetscErrorCode DMPlexMetricSetNoSurf(PetscDM, PetscBool)
    PetscErrorCode DMPlexMetricNoSurf(PetscDM, PetscBool*)
    PetscErrorCode DMPlexMetricSetVerbosity(PetscDM, PetscInt)
    PetscErrorCode DMPlexMetricGetVerbosity(PetscDM, PetscInt*)
    PetscErrorCode DMPlexMetricSetNumIterations(PetscDM, PetscInt)
    PetscErrorCode DMPlexMetricGetNumIterations(PetscDM, PetscInt*)
    PetscErrorCode DMPlexMetricSetMinimumMagnitude(PetscDM, PetscReal)
    PetscErrorCode DMPlexMetricGetMinimumMagnitude(PetscDM, PetscReal*)
    PetscErrorCode DMPlexMetricSetMaximumMagnitude(PetscDM, PetscReal)
    PetscErrorCode DMPlexMetricGetMaximumMagnitude(PetscDM, PetscReal*)
    PetscErrorCode DMPlexMetricSetMaximumAnisotropy(PetscDM, PetscReal)
    PetscErrorCode DMPlexMetricGetMaximumAnisotropy(PetscDM, PetscReal*)
    PetscErrorCode DMPlexMetricSetTargetComplexity(PetscDM, PetscReal)
    PetscErrorCode DMPlexMetricGetTargetComplexity(PetscDM, PetscReal*)
    PetscErrorCode DMPlexMetricSetNormalizationOrder(PetscDM, PetscReal)
    PetscErrorCode DMPlexMetricGetNormalizationOrder(PetscDM, PetscReal*)
    PetscErrorCode DMPlexMetricSetGradationFactor(PetscDM, PetscReal)
    PetscErrorCode DMPlexMetricGetGradationFactor(PetscDM, PetscReal*)
    PetscErrorCode DMPlexMetricSetHausdorffNumber(PetscDM, PetscReal)
    PetscErrorCode DMPlexMetricGetHausdorffNumber(PetscDM, PetscReal*)
    PetscErrorCode DMPlexMetricCreate(PetscDM, PetscInt, PetscVec*)
    PetscErrorCode DMPlexMetricCreateUniform(PetscDM, PetscInt, PetscReal, PetscVec*)
    PetscErrorCode DMPlexMetricCreateIsotropic(PetscDM, PetscInt, PetscVec, PetscVec*)
    PetscErrorCode DMPlexMetricDeterminantCreate(PetscDM, PetscInt, PetscVec*, PetscDM*)
    PetscErrorCode DMPlexMetricEnforceSPD(PetscDM, PetscVec, PetscBool, PetscBool, PetscVec, PetscVec)
    PetscErrorCode DMPlexMetricNormalize(PetscDM, PetscVec, PetscBool, PetscBool, PetscVec, PetscVec)
    PetscErrorCode DMPlexMetricAverage2(PetscDM, PetscVec, PetscVec, PetscVec)
    PetscErrorCode DMPlexMetricAverage3(PetscDM, PetscVec, PetscVec, PetscVec, PetscVec)
    PetscErrorCode DMPlexMetricIntersection2(PetscDM, PetscVec, PetscVec, PetscVec)
    PetscErrorCode DMPlexMetricIntersection3(PetscDM, PetscVec, PetscVec, PetscVec, PetscVec)

    PetscErrorCode DMPlexComputeGradientClementInterpolant(PetscDM, PetscVec, PetscVec)

    PetscErrorCode DMPlexTopologyView(PetscDM, PetscViewer)
    PetscErrorCode DMPlexCoordinatesView(PetscDM, PetscViewer)
    PetscErrorCode DMPlexLabelsView(PetscDM, PetscViewer)
    PetscErrorCode DMPlexSectionView(PetscDM, PetscViewer, PetscDM)
    PetscErrorCode DMPlexGlobalVectorView(PetscDM, PetscViewer, PetscDM, PetscVec)
    PetscErrorCode DMPlexLocalVectorView(PetscDM, PetscViewer, PetscDM, PetscVec)

    PetscErrorCode DMPlexTopologyLoad(PetscDM, PetscViewer, PetscSF*)
    PetscErrorCode DMPlexCoordinatesLoad(PetscDM, PetscViewer, PetscSF)
    PetscErrorCode DMPlexLabelsLoad(PetscDM, PetscViewer, PetscSF)
    PetscErrorCode DMPlexSectionLoad(PetscDM, PetscViewer, PetscDM, PetscSF, PetscSF*, PetscSF*)
    PetscErrorCode DMPlexGlobalVectorLoad(PetscDM, PetscViewer, PetscDM, PetscSF, PetscVec)
    PetscErrorCode DMPlexLocalVectorLoad(PetscDM, PetscViewer, PetscDM, PetscSF, PetscVec)

    PetscErrorCode DMPlexTransformApply(PetscDMPlexTransform, PetscDM, PetscDM *)
    PetscErrorCode DMPlexTransformCreate(MPI_Comm, PetscDMPlexTransform *)
    PetscErrorCode DMPlexTransformDestroy(PetscDMPlexTransform*)
    PetscErrorCode DMPlexTransformGetType(PetscDMPlexTransform, PetscDMPlexTransformType *)
    PetscErrorCode DMPlexTransformSetType(PetscDMPlexTransform tr, PetscDMPlexTransformType method)
    PetscErrorCode DMPlexTransformSetFromOptions(PetscDMPlexTransform)
    PetscErrorCode DMPlexTransformSetDM(PetscDMPlexTransform, PetscDM)
    PetscErrorCode DMPlexTransformSetUp(PetscDMPlexTransform)
    PetscErrorCode DMPlexTransformView(PetscDMPlexTransform tr, PetscViewer v)

    PetscErrorCode DMPlexCreateNaturalVector(PetscDM, PetscVec *)
    PetscErrorCode DMPlexNaturalToGlobalBegin(PetscDM, PetscVec, PetscVec)
    PetscErrorCode DMPlexNaturalToGlobalEnd(PetscDM, PetscVec, PetscVec)
    PetscErrorCode DMPlexGlobalToNaturalBegin(PetscDM, PetscVec, PetscVec)
    PetscErrorCode DMPlexGlobalToNaturalEnd(PetscDM, PetscVec, PetscVec)
    PetscErrorCode DMPlexGetMigrationSF(PetscDM, PetscSF *)
    PetscErrorCode DMPlexSetMigrationSF(PetscDM, PetscSF)
    PetscErrorCode DMPlexCreateGlobalToNaturalSF(PetscDM, PetscSection, PetscSF, PetscSF *)
    PetscErrorCode DMPlexMigrateGlobalToNaturalSF(PetscDM, PetscDM, PetscSF, PetscSF, PetscSF *)



from petsc4py import PETSc
import unittest

import numpy as np


def mkdata(comm, m, N, bs):
    start = m * comm.rank
    end = start + m
    idt = PETSc.IntType
    sdt = PETSc.ScalarType
    rows = np.array(range(start, end), dtype=idt)
    cols = np.array(range(N), dtype=idt)
    vals = np.array(range(m * N * bs * bs), dtype=sdt)
    vals.shape = (-1, bs, bs)
    return rows, cols, vals


class BaseTestMatAnyDense:
    COMM = PETSc.COMM_NULL
    GRID = 0, 0
    BSIZE = None
    TYPE = PETSc.Mat.Type.DENSE

    def setUp(self):
        COMM = self.COMM
        GM, GN = self.GRID
        BS = self.BSIZE  # or 1
        #
        self.A = PETSc.Mat().create(comm=COMM)
        bs = BS or 1
        m, N = GM, GN
        rowsz = (m * bs, None)
        colsz = (None, N * bs)
        self.A.setSizes([rowsz, colsz], BS)
        self.A.setType(self.TYPE)

    def tearDown(self):
        self.A.destroy()
        self.A = None
        PETSc.garbage_cleanup()

    def testSetValues(self):
        self._preallocate()
        r, c, v = self._set_values()
        self.A.assemble()
        self._chk_array(self.A, r, c, v)
        r, c, v = self._set_values()
        self.A.assemble()
        self._chk_array(self.A, r, c, v)

    def testGetDiagonalBlock(self):
        M, N = self.A.getSize()
        # only for square matrices
        if M != N:
            return
        self._preallocate()
        self._set_values()
        self.A.assemble()
        B = self.A.getDiagonalBlock()
        self.assertEqual(self.A.getLocalSize(), B.getSize())
        B.destroy()

    def testSubMatrix(self):
        self._preallocate()
        self._set_values()
        self.A.assemble()
        B = self.A.getDenseSubMatrix()
        X = B.copy()
        self.A.restoreDenseSubMatrix(B)
        self.assertTrue(self.A.equal(X))
        X.destroy()
        M, N = self.A.getSize()
        rst = min(1,M)
        cst = min(2,N)
        ren = min(3,M)
        cen = min(5,N)
        B = self.A.getDenseSubMatrix(rst, ren, cst, cen)
        self.assertTrue(B.getSize(), (ren - rst, cen - cst))
        self.A.restoreDenseSubMatrix(B)
        self.assertFalse(B)
        B = self.A.getDenseSubMatrix(cbegin=cst, rbegin=rst, cend=cen, rend=ren)
        self.assertTrue(B.getSize(), (ren - rst, cen - cst))
        self.A.restoreDenseSubMatrix(B)
        self.assertFalse(B)

    def testColumnVec(self):
        self._preallocate()
        self._set_values()
        self.A.assemble()
        x = self.A.createVecLeft()
        x.setRandom()
        for i in range(self.A.getSize()[1]):
            c = self.A.getDenseColumnVec(i)
            x.copy(c)
            self.A.restoreDenseColumnVec(i, V=c)
            self.assertFalse(c)
            c = self.A.getDenseColumnVec(i,'r')
            self.assertTrue(x.equal(c))
            self.A.restoreDenseColumnVec(i,'r', c)
            self.assertFalse(c)
            c = self.A.getDenseColumnVec(i,'w')
            c.set(0)
            self.A.restoreDenseColumnVec(i,'w', c)
            self.assertFalse(c)
            if i > 0:
                c = self.A.getDenseColumnVec(i-1, mode='r')
                self.assertEqual(c.norm(), 0)
                self.A.restoreDenseColumnVec(i-1, V=c, mode='r')
                self.assertFalse(c)
        x.destroy()

    def testCreateTranspose(self):
        self._preallocate()
        self._set_values()
        self.A.assemble()
        A = self.A
        AT = PETSc.Mat().createTranspose(A)
        x, y = A.createVecs()
        xt, yt = AT.createVecs()
        #
        y.setRandom()
        A.multTranspose(y, x)
        y.copy(xt)
        AT.mult(xt, yt)
        self.assertTrue(yt.equal(x))
        #
        x.setRandom()
        A.mult(x, y)
        x.copy(yt)
        AT.multTranspose(yt, xt)
        self.assertTrue(xt.equal(y))
        #
        underlyingA = AT.getTransposeMat()
        self.assertTrue(underlyingA.equal(A))

    def _preallocate(self):
        self.A.setPreallocationDense(None)

    def _set_values(self):
        COMM = self.COMM
        GM, GN = self.GRID
        BS = self.BSIZE or 1
        rows, cols, vals = mkdata(COMM, GM, GN, BS)
        if not self.BSIZE:
            setvalues = self.A.setValues
        else:
            setvalues = self.A.setValuesBlocked
        setvalues(rows, cols, vals)
        return rows, cols, vals

    def _chk_bs(self, A, bs):
        self.assertEqual(A.getBlockSize(), bs or 1)

    def _chk_array(self, A, r, c, v):
        return  # XXX
        vals = self.A.getValues(r, c)
        vals.shape = v.shape
        self.assertTrue(np.allclose(vals, v))


# -- Dense ---------------------


class BaseTestMatDense(BaseTestMatAnyDense, unittest.TestCase):
    COMM = PETSc.COMM_WORLD
    GRID = 0, 0
    BSIZE = None


# -- Seq Dense --


class TestMatSeqDense(BaseTestMatDense):
    COMM = PETSc.COMM_SELF
    TYPE = PETSc.Mat.Type.SEQDENSE


class TestMatSeqDense_G23(TestMatSeqDense):
    GRID = 2, 3


class TestMatSeqDense_G45(TestMatSeqDense):
    GRID = 4, 5


class TestMatSeqDense_G77(TestMatSeqDense):
    GRID = 7, 7


class TestMatSeqDense_G89(TestMatSeqDense):
    GRID = 8, 9


# -- MPI Dense --


class TestMatMPIDense(BaseTestMatDense):
    COMM = PETSc.COMM_WORLD
    TYPE = PETSc.Mat.Type.MPIDENSE


class TestMatMPIDense_G23(TestMatMPIDense):
    GRID = 2, 3


class TestMatMPIDense_G45(TestMatMPIDense):
    GRID = 4, 5


class TestMatMPIDense_G77(TestMatMPIDense):
    GRID = 7, 7


class TestMatMPIDense_G89(TestMatMPIDense):
    GRID = 8, 9


# -- Dense + Block ---------------


class BaseTestMatDense_B(BaseTestMatAnyDense, unittest.TestCase):
    COMM = PETSc.COMM_WORLD
    GRID = 0, 0
    BSIZE = 1

    def _preallocate(self):
        # self.A.setBlockSize(self.BSIZE)
        self.A.setPreallocationDense(None)
        # self.A.setBlockSize(self.BSIZE)
        self._chk_bs(self.A, self.BSIZE)


# -- Seq Dense + Block --


class TestMatSeqDense_B(BaseTestMatDense_B):
    COMM = PETSc.COMM_SELF
    TYPE = PETSc.Mat.Type.SEQDENSE


# bs = 1
class TestMatSeqDense_B_G23(TestMatSeqDense_B):
    GRID = 2, 3


class TestMatSeqDense_B_G45(TestMatSeqDense_B):
    GRID = 4, 5


class TestMatSeqDense_B_G89(TestMatSeqDense_B):
    GRID = 8, 9


# bs = 2
class TestMatSeqDense_B_G23_B2(TestMatSeqDense_B_G23):
    BSIZE = 2


class TestMatSeqDense_B_G45_B2(TestMatSeqDense_B_G45):
    BSIZE = 2


class TestMatSeqDense_B_G89_B2(TestMatSeqDense_B_G89):
    BSIZE = 2


# bs = 3
class TestMatSeqDense_B_G23_B3(TestMatSeqDense_B_G23):
    BSIZE = 3


class TestMatSeqDense_B_G45_B3(TestMatSeqDense_B_G45):
    BSIZE = 3


class TestMatSeqDense_B_G89_B3(TestMatSeqDense_B_G89):
    BSIZE = 3


# bs = 4
class TestMatSeqDense_B_G23_B4(TestMatSeqDense_B_G23):
    BSIZE = 4


class TestMatSeqDense_B_G45_B4(TestMatSeqDense_B_G45):
    BSIZE = 4


class TestMatSeqDense_B_G89_B4(TestMatSeqDense_B_G89):
    BSIZE = 4


# bs = 5
class TestMatSeqDense_B_G23_B5(TestMatSeqDense_B_G23):
    BSIZE = 5


class TestMatSeqDense_B_G45_B5(TestMatSeqDense_B_G45):
    BSIZE = 5


class TestMatSeqDense_B_G89_B5(TestMatSeqDense_B_G89):
    BSIZE = 5


# -- MPI Dense + Block --


class TestMatMPIDense_B(BaseTestMatDense_B):
    COMM = PETSc.COMM_WORLD
    TYPE = PETSc.Mat.Type.MPIDENSE


# bs = 1
class TestMatMPIDense_B_G23(TestMatMPIDense_B):
    GRID = 2, 3


class TestMatMPIDense_B_G45(TestMatMPIDense_B):
    GRID = 4, 5


class TestMatMPIDense_B_G77(TestMatMPIDense_B):
    GRID = 7, 7


class TestMatMPIDense_B_G89(TestMatMPIDense_B):
    GRID = 8, 9


# bs = 2
class TestMatMPIDense_B_G23_B2(TestMatMPIDense_B_G23):
    BSIZE = 2


class TestMatMPIDense_B_G45_B2(TestMatMPIDense_B_G45):
    BSIZE = 2


class TestMatMPIDense_B_G77_B2(TestMatMPIDense_B_G77):
    BSIZE = 2


class TestMatMPIDense_B_G89_B2(TestMatMPIDense_B_G89):
    BSIZE = 2


# bs = 3
class TestMatMPIDense_B_G23_B3(TestMatMPIDense_B_G23):
    BSIZE = 3


class TestMatMPIDense_B_G45_B3(TestMatMPIDense_B_G45):
    BSIZE = 3


class TestMatMPIDense_B_G77_B3(TestMatMPIDense_B_G77):
    BSIZE = 3


class TestMatMPIDense_B_G89_B3(TestMatMPIDense_B_G89):
    BSIZE = 3


# bs = 4
class TestMatMPIDense_B_G23_B4(TestMatMPIDense_B_G23):
    BSIZE = 4


class TestMatMPIDense_B_G45_B4(TestMatMPIDense_B_G45):
    BSIZE = 4


class TestMatMPIDense_B_G77_B4(TestMatMPIDense_B_G77):
    BSIZE = 4


class TestMatMPIDense_B_G89_B4(TestMatMPIDense_B_G89):
    BSIZE = 4


# bs = 5
class TestMatMPIDense_B_G23_B5(TestMatMPIDense_B_G23):
    BSIZE = 5


class TestMatMPIDense_B_G45_B5(TestMatMPIDense_B_G45):
    BSIZE = 5


class TestMatMPIDense_B_G77_B5(TestMatMPIDense_B_G77):
    BSIZE = 5


class TestMatMPIDense_B_G89_B5(TestMatMPIDense_B_G89):
    BSIZE = 5


# -----

if __name__ == '__main__':
    unittest.main()



cdef set appctx_registry = set()

cdef inline object registerAppCtx(void *appctx):
    cdef object key = <Py_uintptr_t> appctx
    appctx_registry.add(key)

cdef inline object toAppCtx(void *appctx):
    cdef object key = <Py_uintptr_t> appctx
    if key in appctx_registry:
        return <object> appctx
    else:
        if appctx != NULL:
            return PyLong_FromVoidPtr(appctx)
        else:
            return None



ctypedef PetscErrorCode (*PetscDMShellXToYFunction)(PetscDM,
                                                    PetscVec,
                                                    PetscInsertMode,
                                                    PetscVec) except PETSC_ERR_PYTHON
cdef extern from * nogil:
    ctypedef PetscErrorCode (*PetscDMShellCreateVectorFunction)(PetscDM,
                                                                PetscVec*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellCreateMatrixFunction)(PetscDM,
                                                                PetscMat*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellTransferFunction)(PetscDM,
                                                            MPI_Comm,
                                                            PetscDM*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellCreateInterpolationFunction)(PetscDM,
                                                                       PetscDM,
                                                                       PetscMat*,
                                                                       PetscVec*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellCreateInjectionFunction)(PetscDM,
                                                                   PetscDM,
                                                                   PetscMat*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellCreateRestrictionFunction)(PetscDM,
                                                                     PetscDM,
                                                                     PetscMat*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellCreateFieldDecompositionFunction)(PetscDM,
                                                                            PetscInt*,
                                                                            char***,
                                                                            PetscIS**,
                                                                            PetscDM**) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellCreateDomainDecompositionFunction)(PetscDM,
                                                                             PetscInt*,
                                                                             char***,
                                                                             PetscIS**,
                                                                             PetscIS**,
                                                                             PetscDM**) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellCreateDomainDecompositionScattersFunction)(PetscDM,
                                                                                     PetscInt,
                                                                                     PetscDM*,
                                                                                     PetscScatter**,
                                                                                     PetscScatter**,
                                                                                     PetscScatter**) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscDMShellCreateSubDM)(PetscDM,
                                                       PetscInt,
                                                       PetscInt[],
                                                       PetscIS*,
                                                       PetscDM*) except PETSC_ERR_PYTHON
    PetscErrorCode DMShellCreate(MPI_Comm, PetscDM*)
    PetscErrorCode DMShellSetMatrix(PetscDM, PetscMat)
    PetscErrorCode DMShellSetGlobalVector(PetscDM, PetscVec)
    PetscErrorCode DMShellSetLocalVector(PetscDM, PetscVec)
    PetscErrorCode DMShellSetCreateGlobalVector(PetscDM, PetscDMShellCreateVectorFunction)
    PetscErrorCode DMShellSetCreateLocalVector(PetscDM, PetscDMShellCreateVectorFunction)
    PetscErrorCode DMShellSetGlobalToLocal(PetscDM, PetscDMShellXToYFunction, PetscDMShellXToYFunction)
    PetscErrorCode DMShellSetGlobalToLocalVecScatter(PetscDM, PetscScatter)
    PetscErrorCode DMShellSetLocalToGlobal(PetscDM, PetscDMShellXToYFunction, PetscDMShellXToYFunction)
    PetscErrorCode DMShellSetLocalToGlobalVecScatter(PetscDM, PetscScatter)
    PetscErrorCode DMShellSetLocalToLocal(PetscDM, PetscDMShellXToYFunction, PetscDMShellXToYFunction)
    PetscErrorCode DMShellSetLocalToLocalVecScatter(PetscDM, PetscScatter)
    PetscErrorCode DMShellSetCreateMatrix(PetscDM, PetscDMShellCreateMatrixFunction)
    PetscErrorCode DMShellSetCoarsen(PetscDM, PetscDMShellTransferFunction)
    PetscErrorCode DMShellSetRefine(PetscDM, PetscDMShellTransferFunction)
    PetscErrorCode DMShellSetCreateInterpolation(PetscDM, PetscDMShellCreateInterpolationFunction)
    PetscErrorCode DMShellSetCreateInjection(PetscDM, PetscDMShellCreateInjectionFunction)
    PetscErrorCode DMShellSetCreateRestriction(PetscDM, PetscDMShellCreateRestrictionFunction)
    PetscErrorCode DMShellSetCreateFieldDecomposition(PetscDM, PetscDMShellCreateFieldDecompositionFunction)
    PetscErrorCode DMShellSetCreateDomainDecomposition(PetscDM, PetscDMShellCreateDomainDecompositionFunction)
    PetscErrorCode DMShellSetCreateDomainDecompositionScatters(PetscDM, PetscDMShellCreateDomainDecompositionScattersFunction)
    PetscErrorCode DMShellSetCreateSubDM(PetscDM, PetscDMShellCreateSubDM)

    PetscErrorCode VecGetDM(PetscVec, PetscDM*)
    PetscErrorCode VecSetDM(PetscVec, PetscDM)
    PetscErrorCode MatGetDM(PetscMat, PetscDM*)
    PetscErrorCode MatSetDM(PetscMat, PetscDM)


cdef PetscErrorCode DMSHELL_CreateGlobalVector(
    PetscDM dm,
    PetscVec *v) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Vec vec
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__create_global_vector__')
    assert context is not None and type(context) is tuple
    (create_gvec, args, kargs) = context
    vec = create_gvec(Dm, *args, **kargs)
    CHKERR(PetscINCREF(vec.obj))
    v[0] = vec.vec
    cdef PetscDM odm = NULL
    CHKERR(VecGetDM(v[0], &odm))
    if odm == NULL:
        CHKERR(VecSetDM(v[0], dm))
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateLocalVector(
    PetscDM dm,
    PetscVec *v) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Vec vec
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__create_local_vector__')
    assert context is not None and type(context) is tuple
    (create_lvec, args, kargs) = context
    vec = create_lvec(Dm, *args, **kargs)
    CHKERR(PetscINCREF(vec.obj))
    v[0] = vec.vec
    cdef PetscDM odm = NULL
    CHKERR(VecGetDM(v[0], &odm))
    if odm == NULL:
        CHKERR(VecSetDM(v[0], dm))
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_GlobalToLocalBegin(
    PetscDM dm,
    PetscVec g,
    PetscInsertMode mode,
    PetscVec l) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Vec gvec = ref_Vec(g)
    cdef Vec lvec = ref_Vec(l)
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__g2l_begin__')
    assert context is not None and type(context) is tuple
    (begin, args, kargs) = context
    begin(Dm, gvec, mode, lvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_GlobalToLocalEnd(
    PetscDM dm,
    PetscVec g,
    PetscInsertMode mode,
    PetscVec l) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Vec gvec = ref_Vec(g)
    cdef Vec lvec = ref_Vec(l)
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__g2l_end__')
    assert context is not None and type(context) is tuple
    (end, args, kargs) = context
    end(Dm, gvec, mode, lvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_LocalToGlobalBegin(
    PetscDM dm,
    PetscVec g,
    PetscInsertMode mode,
    PetscVec l) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Vec gvec = ref_Vec(g)
    cdef Vec lvec = ref_Vec(l)
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__l2g_begin__')
    assert context is not None and type(context) is tuple
    (begin, args, kargs) = context
    begin(Dm, gvec, mode, lvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_LocalToGlobalEnd(
    PetscDM dm,
    PetscVec g,
    PetscInsertMode mode,
    PetscVec l) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Vec gvec = ref_Vec(g)
    cdef Vec lvec = ref_Vec(l)
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__l2g_end__')
    assert context is not None and type(context) is tuple
    (end, args, kargs) = context
    end(Dm, gvec, mode, lvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_LocalToLocalBegin(
    PetscDM dm,
    PetscVec g,
    PetscInsertMode mode,
    PetscVec l) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Vec gvec = ref_Vec(g)
    cdef Vec lvec = ref_Vec(l)
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__l2l_begin__')
    assert context is not None and type(context) is tuple
    (begin, args, kargs) = context
    begin(Dm, gvec, mode, lvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_LocalToLocalEnd(
    PetscDM dm,
    PetscVec g,
    PetscInsertMode mode,
    PetscVec l) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Vec gvec = ref_Vec(g)
    cdef Vec lvec = ref_Vec(l)
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__l2l_end__')
    assert context is not None and type(context) is tuple
    (end, args, kargs) = context
    end(Dm, gvec, mode, lvec, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateMatrix(
    PetscDM dm,
    PetscMat *cmat) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef Mat mat
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__create_matrix__')
    assert context is not None and type(context) is tuple
    (matrix, args, kargs) = context
    mat = matrix(Dm, *args, **kargs)
    CHKERR(PetscINCREF(mat.obj))
    cmat[0] = mat.mat
    cdef PetscDM odm = NULL
    CHKERR(MatGetDM(cmat[0], &odm))
    if odm == NULL:
        CHKERR(MatSetDM(cmat[0], dm))
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_Coarsen(
    PetscDM dm,
    MPI_Comm comm,
    PetscDM *dmc) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef DM Dmc
    cdef Comm Comm = new_Comm(comm)
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__coarsen__')
    assert context is not None and type(context) is tuple
    (coarsen, args, kargs) = context
    Dmc = coarsen(Dm, Comm, *args, **kargs)
    CHKERR(PetscINCREF(Dmc.obj))
    dmc[0] = Dmc.dm
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_Refine(
    PetscDM dm,
    MPI_Comm comm,
    PetscDM *dmf) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef DM Dmf
    cdef Comm Comm = new_Comm(comm)
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__refine__')
    assert context is not None and type(context) is tuple
    (refine, args, kargs) = context
    Dmf = refine(Dm, Comm, *args, **kargs)
    CHKERR(PetscINCREF(Dmf.obj))
    dmf[0] = Dmf.dm
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateInterpolation(
    PetscDM dmc,
    PetscDM dmf,
    PetscMat *cmat,
    PetscVec *cvec) except PETSC_ERR_PYTHON with gil:
    cdef DM Dmc = subtype_DM(dmc)()
    cdef DM Dmf = subtype_DM(dmf)()
    cdef Mat mat
    cdef Vec vec
    Dmc.dm = dmc
    CHKERR(PetscINCREF(Dmc.obj))
    Dmf.dm = dmf
    CHKERR(PetscINCREF(Dmf.obj))
    context = Dmc.get_attr('__create_interpolation__')
    assert context is not None and type(context) is tuple
    (interpolation, args, kargs) = context
    mat, vec = interpolation(Dmc, Dmf, *args, **kargs)
    CHKERR(PetscINCREF(mat.obj))
    cmat[0] = mat.mat
    if cvec == NULL:
        return PETSC_SUCCESS
    if vec is None:
        cvec[0] = NULL
    else:
        CHKERR(PetscINCREF(vec.obj))
        cvec[0] = vec.vec
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateInjection(
    PetscDM dmc,
    PetscDM dmf,
    PetscMat *cmat) except PETSC_ERR_PYTHON with gil:
    cdef DM Dmc = subtype_DM(dmc)()
    cdef DM Dmf = subtype_DM(dmf)()
    cdef Mat mat
    Dmc.dm = dmc
    CHKERR(PetscINCREF(Dmc.obj))
    Dmf.dm = dmf
    CHKERR(PetscINCREF(Dmf.obj))
    context = Dmc.get_attr('__create_injection__')
    assert context is not None and type(context) is tuple
    (injection, args, kargs) = context
    mat = injection(Dmc, Dmf, *args, **kargs)
    CHKERR(PetscINCREF(mat.obj))
    cmat[0] = mat.mat
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateRestriction(
    PetscDM dmf,
    PetscDM dmc,
    PetscMat *cmat) except PETSC_ERR_PYTHON with gil:
    cdef DM Dmf = subtype_DM(dmf)()
    cdef DM Dmc = subtype_DM(dmc)()
    cdef Mat mat
    Dmf.dm = dmf
    CHKERR(PetscINCREF(Dmf.obj))
    Dmc.dm = dmc
    CHKERR(PetscINCREF(Dmc.obj))
    context = Dmf.get_attr('__create_restriction__')
    assert context is not None and type(context) is tuple
    (restriction, args, kargs) = context
    mat = restriction(Dmf, Dmc, *args, **kargs)
    CHKERR(PetscINCREF(mat.obj))
    cmat[0] = mat.mat
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateFieldDecomposition(
    PetscDM dm,
    PetscInt *clen,
    char ***namelist,
    PetscIS **islist,
    PetscDM **dmlist) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef int i
    cdef const char *cname = NULL
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__create_field_decomp__')
    assert context is not None and type(context) is tuple
    (decomp, args, kargs) = context
    names, ises, dms = decomp(Dm, *args, **kargs)

    if clen != NULL:
        if names is not None:
            clen[0] = <PetscInt>len(names)
        elif ises is not None:
            clen[0] = <PetscInt>len(ises)
        elif dms is not None:
            clen[0] = <PetscInt>len(dms)
        else:
            clen[0] = 0

    if namelist != NULL and names is not None:
        CHKERR(PetscMalloc(len(names)*sizeof(char**), namelist))
        for i in range(len(names)):
            names[i] = str2bytes(names[i], &cname)
            CHKERR(PetscStrallocpy(cname, &namelist[0][i]))

    if islist != NULL and ises is not None:
        CHKERR(PetscMalloc(len(ises)*sizeof(PetscIS), islist))
        for i in range(len(ises)):
            islist[0][i] = (<IS?>ises[i]).iset
            CHKERR(PetscINCREF((<IS?>ises[i]).obj))

    if dmlist != NULL and dms is not None:
        CHKERR(PetscMalloc(len(dms)*sizeof(PetscDM), dmlist))
        for i in range(len(dms)):
            dmlist[0][i] = (<DM?>dms[i]).dm
            CHKERR(PetscINCREF((<DM?>dms[i]).obj))
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateDomainDecomposition(
    PetscDM dm,
    PetscInt *clen,
    char ***namelist,
    PetscIS **innerislist,
    PetscIS **outerislist,
    PetscDM **dmlist) except PETSC_ERR_PYTHON with gil:
    cdef DM Dm = subtype_DM(dm)()
    cdef int i
    cdef const char *cname = NULL
    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))
    context = Dm.get_attr('__create_domain_decomp__')
    assert context is not None and type(context) is tuple
    (decomp, args, kargs) = context
    names, innerises, outerises, dms = decomp(Dm, *args, **kargs)

    if clen != NULL:
        if names is not None:
            clen[0] = <PetscInt>len(names)
        elif innerises is not None:
            clen[0] = <PetscInt>len(innerises)
        elif outerises is not None:
            clen[0] = <PetscInt>len(outerises)
        elif dms is not None:
            clen[0] = <PetscInt>len(dms)
        else:
            clen[0] = 0

    if namelist != NULL and names is not None:
        CHKERR(PetscMalloc(len(names)*sizeof(char**), namelist))
        for i in range(len(names)):
            names[i] = str2bytes(names[i], &cname)
            CHKERR(PetscStrallocpy(cname, &namelist[0][i]))

    if innerislist != NULL and innerises is not None:
        CHKERR(PetscMalloc(len(innerises)*sizeof(PetscIS), innerislist))
        for i in range(len(innerises)):
            innerislist[0][i] = (<IS?>innerises[i]).iset
            CHKERR(PetscINCREF((<IS?>innerises[i]).obj))

    if outerislist != NULL and outerises is not None:
        CHKERR(PetscMalloc(len(outerises)*sizeof(PetscIS), outerislist))
        for i in range(len(outerises)):
            outerislist[0][i] = (<IS?>outerises[i]).iset
            CHKERR(PetscINCREF((<IS?>outerises[i]).obj))

    if dmlist != NULL and dms is not None:
        CHKERR(PetscMalloc(len(dms)*sizeof(PetscDM), dmlist))
        for i in range(len(dms)):
            dmlist[0][i] = (<DM?>dms[i]).dm
            CHKERR(PetscINCREF((<DM?>dms[i]).obj))
    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateDomainDecompositionScatters(
    PetscDM dm,
    PetscInt clen,
    PetscDM *subdms,
    PetscScatter** iscat,
    PetscScatter** oscat,
    PetscScatter** gscat) except PETSC_ERR_PYTHON with gil:

    cdef DM Dm = subtype_DM(dm)()
    cdef int i
    cdef DM subdm = None

    Dm.dm = dm
    CHKERR(PetscINCREF(Dm.obj))

    psubdms = []
    for i from 0 <= i < clen:
        subdm = subtype_DM(subdms[i])()
        subdm.dm = subdms[i]
        CHKERR(PetscINCREF(subdm.obj))
        psubdms.append(subdm)

    context = Dm.get_attr('__create_domain_decomp_scatters__')
    assert context is not None and type(context) is tuple
    (scatters, args, kargs) = context
    (iscatter, oscatter, gscatter) = scatters(Dm, psubdms, *args, **kargs)

    assert len(iscatter) == clen
    assert len(oscatter) == clen
    assert len(gscatter) == clen

    CHKERR (PetscMalloc(clen*sizeof(PetscScatter), iscat))
    CHKERR (PetscMalloc(clen*sizeof(PetscScatter), oscat))
    CHKERR (PetscMalloc(clen*sizeof(PetscScatter), gscat))

    for i in range(clen):
        iscat[0][i] = (<Scatter?>iscatter[i]).sct
        CHKERR(PetscINCREF((<Scatter?>iscatter[i]).obj))

        oscat[0][i] = (<Scatter?>oscatter[i]).sct
        CHKERR(PetscINCREF((<Scatter?>oscatter[i]).obj))

        gscat[0][i] = (<Scatter?>gscatter[i]).sct
        CHKERR(PetscINCREF((<Scatter?>gscatter[i]).obj))

    return PETSC_SUCCESS

cdef PetscErrorCode DMSHELL_CreateSubDM(
    PetscDM cdm,
    PetscInt numFields,
    const PetscInt cfields[],
    PetscIS *ciset,
    PetscDM *csubdm) except PETSC_ERR_PYTHON with gil:
    cdef DM dm = subtype_DM(cdm)()
    cdef IS iset
    cdef DM subdm
    dm.dm = cdm
    CHKERR(PetscINCREF(dm.obj))
    context = dm.get_attr('__create_subdm__')
    assert context is not None and type(context) is tuple
    (create_subdm, args, kargs) = context

    fields = array_i(numFields, cfields)

    iset, subdm = create_subdm(dm, fields, *args, **kargs)

    CHKERR(PetscINCREF(iset.obj))
    CHKERR(PetscINCREF(subdm.obj))
    ciset[0] = iset.iset
    csubdm[0] = subdm.dm
    return PETSC_SUCCESS



# --------------------------------------------------------------------

class MatPartitioningType(object):
    """The partitioning types."""
    PARTITIONINGCURRENT  = S_(MATPARTITIONINGCURRENT)
    PARTITIONINGAVERAGE  = S_(MATPARTITIONINGAVERAGE)
    PARTITIONINGSQUARE   = S_(MATPARTITIONINGSQUARE)
    PARTITIONINGPARMETIS = S_(MATPARTITIONINGPARMETIS)
    PARTITIONINGCHACO    = S_(MATPARTITIONINGCHACO)
    PARTITIONINGPARTY    = S_(MATPARTITIONINGPARTY)
    PARTITIONINGPTSCOTCH = S_(MATPARTITIONINGPTSCOTCH)
    PARTITIONINGHIERARCH = S_(MATPARTITIONINGHIERARCH)

# --------------------------------------------------------------------


cdef class MatPartitioning(Object):
    """Object for managing the partitioning of a matrix or graph."""

    Type = MatPartitioningType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.part
        self.part = NULL

    def __call__(self):
        return self.getValue()

    def view(self, Viewer viewer=None) -> None:
        """View the partitioning data structure.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` to display the graph.

        See Also
        --------
        petsc.MatPartitioningView

        """
        assert self.obj != NULL
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(MatPartitioningView(self.part, vwr))

    def destroy(self) -> Self:
        """Destroy the partitioning context.

        Collective.

        See Also
        --------
        create, petsc.MatPartitioningDestroy

        """
        CHKERR(MatPartitioningDestroy(&self.part))
        return self

    def create(self, comm: Comm | None = None) -> Self:
        """Create a partitioning context.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        destroy, petsc.MatPartitioningCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        CHKERR(MatPartitioningCreate(ccomm, &self.part))
        return self

    def setType(self, matpartitioning_type: Type | str) -> None:
        """Set the type of the partitioner to use.

        Collective.

        Parameters
        ----------
        matpartitioning_type
            The partitioner type.

        See Also
        --------
        getType, petsc.MatPartitioningSetType

        """
        cdef PetscMatPartitioningType cval = NULL
        matpartitioning_type = str2bytes(matpartitioning_type, &cval)
        CHKERR(MatPartitioningSetType(self.part, cval))

    def getType(self) -> str:
        """Return the partitioning method.

        Not collective.

        See Also
        --------
        setType, petsc.MatPartitioningGetType

        """
        cdef PetscMatPartitioningType cval = NULL
        CHKERR(MatPartitioningGetType(self.part, &cval))
        return bytes2str(cval)

    def setFromOptions(self) -> None:
        """Set parameters in the partitioner from the options database.

        Collective.

        See Also
        --------
        petsc_options, petsc.MatPartitioningSetFromOptions

        """
        CHKERR(MatPartitioningSetFromOptions(self.part))

    def setAdjacency(self, Mat adj) -> None:
        """Set the adjacency graph (matrix) of the thing to be partitioned.

        Collective.

        Parameters
        ----------
        adj
            The adjacency matrix, this can be any `Mat.Type` but the natural
            representation is `Mat.Type.MPIADJ`.

        See Also
        --------
        petsc.MatPartitioningSetAdjacency

        """
        CHKERR(MatPartitioningSetAdjacency(self.part, adj.mat))

    def apply(self, IS partitioning) -> None:
        """Return a partitioning for the graph represented by a sparse matrix.

        Collective.

        For each local node this tells the processor number that that node is
        assigned to.

        See Also
        --------
        petsc.MatPartitioningApply

        """
        CHKERR(MatPartitioningApply(self.part, &partitioning.iset))

# --------------------------------------------------------------------

del MatPartitioningType

# --------------------------------------------------------------------



import sys, petsc4py
petsc4py.init(sys.argv)

from petsc4py import PETSc
import Bratu2D as Bratu2D

class App(object):

    def __init__(self, da, lambda_):
        assert da.getDim() == 2
        self.da = da
        self.lambda_ = lambda_

    def formInitGuess(self, snes, X):
        X.zeroEntries() # just in case
        da = self.da.fortran
        vec_X = X.fortran
        ierr = Bratu2D.FormInitGuess(da, vec_X, self.lambda_)
        if ierr: raise PETSc.Error(ierr)

    def formFunction(self, snes, X, F):
        F.zeroEntries() # just in case
        da = self.da.fortran
        vec_X = X.fortran
        vec_F = F.fortran
        ierr = Bratu2D.FormFunction(da, vec_X, vec_F, self.lambda_)
        if ierr: raise PETSc.Error(ierr)

    def formJacobian(self, snes, X, J, P):
        P.zeroEntries() # just in case
        da = self.da.fortran
        vec_X = X.fortran
        mat_P = P.fortran
        ierr = Bratu2D.FormJacobian(da, vec_X, mat_P, self.lambda_)
        if ierr: raise PETSc.Error(ierr)
        if J != P: J.assemble() # matrix-free operator
        return PETSc.Mat.Structure.SAME_NONZERO_PATTERN


OptDB = PETSc.Options()

N = OptDB.getInt('N', 16)
lambda_ = OptDB.getReal('lambda', 6.0)
do_plot = OptDB.getBool('plot', False)

da = PETSc.DA().create([N, N], stencil_width=1)
app = App(da, lambda_)

snes = PETSc.SNES().create()
F = da.createGlobalVec()
snes.setFunction(app.formFunction, F)
J = da.createMat()
snes.setJacobian(app.formJacobian, J)

snes.setFromOptions()

X = da.createGlobalVec()
app.formInitGuess(snes, X)
snes.solve(None, X)

U = da.createNaturalVec()
da.globalToNatural(X, U)

def plot(da, U):
    comm = da.getComm()
    scatter, U0 = PETSc.Scatter.toZero(U)
    scatter.scatter(U, U0, False, PETSc.Scatter.Mode.FORWARD)
    rank = comm.getRank()
    if rank == 0:
        solution = U0[...]
        solution = solution.reshape(da.sizes, order='f').copy()
        try:
            from matplotlib import pyplot
            pyplot.contourf(solution)
            pyplot.axis('equal')
            pyplot.show()
        except:
            pass
    comm.barrier()
    scatter.destroy()
    U0.destroy()

if do_plot: plot(da, U)


U.destroy()
X.destroy()
F.destroy()
J.destroy()
da.destroy()
snes.destroy()



# --------------------------------------------------------------------

cdef extern from * nogil:
    int printf(char *, ...)

cdef extern from "Python.h":
    """
    #if defined(Py_LIMITED_API)
    #define _pytype_enable_gc(t, traverse, clear) \
    do { (void)(traverse); (void)(clear); } while (0)
    #else
    #define _pytype_enable_gc(t, traverse, clear) \
    do { (t)->tp_traverse = (traverse); (t)->tp_clear = (clear); } while (0)
    #endif
    #if PY_VERSION_HEX < 0x030B0000 && !defined(Py_Version)
    #define Py_Version __Pyx_get_runtime_version()
    #endif
    """
    const unsigned long Py_Version
    enum: Py_TPFLAGS_HEAPTYPE
    ctypedef struct PyObject
    ctypedef struct PyTypeObject
    void Py_VISIT(void*) noexcept
    PyTypeObject *Py_TYPE(PyObject *) noexcept
    unsigned long PyType_GetFlags(PyTypeObject *type) noexcept
    ctypedef int (*visitproc)(PyObject *, void *) noexcept
    ctypedef int (*traverseproc)(PyObject *, visitproc, void *) noexcept
    ctypedef int (*inquiry)(PyObject *) noexcept
    void _pytype_enable_gc(PyTypeObject *, traverseproc, inquiry)

cdef extern from "<petsc/private/garbagecollector.h>" nogil:
    PetscErrorCode PetscGarbageCleanup(MPI_Comm)
    PetscErrorCode PetscGarbageView(MPI_Comm, PetscViewer)

cdef int tp_traverse(PyObject *o, visitproc _visit, void *_arg) noexcept:
    cdef visitproc visit "visit" = _visit
    cdef void *arg "arg" = _arg
    <void> visit
    <void> arg
    if Py_Version >= 0x03090000:
        if not (PyType_GetFlags(Py_TYPE(o)) & Py_TPFLAGS_HEAPTYPE):
            Py_VISIT(Py_TYPE(o))
    cdef PetscObject p = (<Object>o).obj[0]
    if p == NULL: return 0
    cdef PyObject *d = <PyObject*>p.python_context
    if d == NULL: return 0
    Py_VISIT(d)
    return 0

cdef int tp_clear(PyObject *o) noexcept:
    cdef PetscObject *p = (<Object>o).obj
    PetscDEALLOC(p)
    return 0

cdef inline void TypeEnableGC(PyTypeObject *t) noexcept:
    _pytype_enable_gc(t, tp_traverse, tp_clear)


def garbage_cleanup(comm: Comm | None = None) -> None:
    """Clean up unused PETSc objects.

    Collective.

    Notes
    -----
    If the communicator ``comm`` if not provided or it is `None`,
    then `COMM_WORLD` is used.

    """
    if not (<int>PetscInitializeCalled): return
    if (<int>PetscFinalizeCalled): return
    cdef MPI_Comm ccomm = MPI_COMM_NULL
    if comm is None:
        ccomm = GetComm(COMM_WORLD, MPI_COMM_NULL)
        CHKERR(PetscGarbageCleanup(ccomm))
    else:
        ccomm = GetComm(comm, MPI_COMM_NULL)
        if ccomm == MPI_COMM_NULL:
            raise ValueError("null communicator")
        CHKERR(PetscGarbageCleanup(ccomm))


def garbage_view(comm: Comm | None = None) -> None:
    """Print summary of the garbage PETSc objects.

    Collective.

    Notes
    -----
    Print out garbage summary on each rank of the communicator ``comm``.
    If no communicator is provided then `COMM_WORLD` is used.

    """
    if not (<int>PetscInitializeCalled): return
    if (<int>PetscFinalizeCalled): return
    cdef MPI_Comm ccomm = MPI_COMM_NULL
    if comm is None:
        comm = COMM_WORLD
    ccomm = GetComm(comm, MPI_COMM_NULL)
    if ccomm == MPI_COMM_NULL:
        raise ValueError("null communicator")
    CHKERR(PetscGarbageView(ccomm, NULL))

# --------------------------------------------------------------------



# --------------------------------------------------------------------

class SpaceType(object):
    """The function space types."""
    POLYNOMIAL = S_(PETSCSPACEPOLYNOMIAL)
    PTRIMMED   = S_(PETSCSPACEPTRIMMED)
    TENSOR     = S_(PETSCSPACETENSOR)
    SUM        = S_(PETSCSPACESUM)
    POINT      = S_(PETSCSPACEPOINT)
    SUBSPACE   = S_(PETSCSPACESUBSPACE)
    WXY        = S_(PETSCSPACEWXY)

# --------------------------------------------------------------------


cdef class Space(Object):
    """Function space object."""
    Type = SpaceType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.space
        self.space  = NULL

    def setUp(self) -> None:
        """Construct data structures for the `Space`.

        Collective.

        See Also
        --------
        petsc.PetscSpaceSetUp

        """
        CHKERR(PetscSpaceSetUp(self.space))

    def create(self, comm: Comm | None = None) -> Self:
        """Create an empty `Space` object.

        Collective.

        The type can then be set with `setType`.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.PetscSpaceCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscSpace newsp = NULL
        CHKERR(PetscSpaceCreate(ccomm, &newsp))
        CHKERR(PetscCLEAR(self.obj)); self.space = newsp
        return self

    def destroy(self) -> Self:
        """Destroy the `Space` object.

        Collective.

        See Also
        --------
        petsc.PetscSpaceDestroy

        """
        CHKERR(PetscSpaceDestroy(&self.space))
        return self

    def view(self, Viewer viewer=None) -> None:
        """View a `Space`.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` to display the `Space`.

        See Also
        --------
        petsc.PetscSpaceView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(PetscSpaceView(self.space, vwr))

    def setFromOptions(self) -> None:
        """Set parameters in `Space` from the options database.

        Collective.

        See Also
        --------
        petsc_options, petsc.PetscSpaceSetFromOptions

        """
        CHKERR(PetscSpaceSetFromOptions(self.space))

    def getDimension(self) -> int:
        """Return the number of basis vectors.

        Not collective.

        See Also
        --------
        petsc.PetscSpaceGetDimension

        """
        cdef PetscInt cdim = 0
        CHKERR(PetscSpaceGetDimension(self.space, &cdim))
        return toInt(cdim)

    def getDegree(self) -> tuple[int, int]:
        """Return the polynomial degrees that characterize this space.

        Not collective.

        Returns
        -------
        minDegree : int
            The degree of the largest polynomial space contained in the space.
        maxDegree : int
            The degree of the smallest polynomial space containing the space.

        See Also
        --------
        setDegree, petsc.PetscSpaceGetDegree

        """
        cdef PetscInt cdegmax = 0, cdegmin = 0
        CHKERR(PetscSpaceGetDegree(self.space, &cdegmin, &cdegmax))
        return toInt(cdegmin), toInt(cdegmax)

    def setDegree(self, degree: int | None, maxDegree: int | None) -> None:
        """Set the degree of approximation for this space.

        Logically collective.

        One of ``degree`` and ``maxDegree`` can be `None`.

        Parameters
        ----------
        degree
            The degree of the largest polynomial space contained in the space.
        maxDegree
            The degree of the largest polynomial space containing the space.

        See Also
        --------
        getDegree, petsc.PetscSpaceSetDegree

        """
        cdef PetscInt cdegree = PETSC_DETERMINE
        if degree is not None: cdegree = asInt(degree)
        cdef PetscInt cmaxdegree = PETSC_DETERMINE
        if maxDegree is not None: cmaxdegree = asInt(maxDegree)
        CHKERR(PetscSpaceSetDegree(self.space, cdegree, cmaxdegree))

    def getNumVariables(self) -> int:
        """Return the number of variables for this space.

        Not collective.

        See Also
        --------
        setNumVariables, petsc.PetscSpaceGetNumVariables

        """
        cdef PetscInt cnvars = 0
        CHKERR(PetscSpaceGetNumVariables(self.space, &cnvars))
        return toInt(cnvars)

    def setNumVariables(self, n: int) -> None:
        """Set the number of variables for this space.

        Logically collective.

        Parameters
        ----------
        n
            The number of variables (``x``, ``y``, ``z`` etc.).

        See Also
        --------
        getNumVariables, petsc.PetscSpaceSetNumVariables

        """
        cdef PetscInt cn = asInt(n)
        CHKERR(PetscSpaceSetNumVariables(self.space, cn))

    def getNumComponents(self) -> int:
        """Return the number of components for this space.

        Not collective.

        See Also
        --------
        setNumComponents, petsc.PetscSpaceGetNumComponents

        """
        cdef PetscInt cncomps = 0
        CHKERR(PetscSpaceGetNumComponents(self.space, &cncomps))
        return toInt(cncomps)

    def setNumComponents(self, nc: int) -> None:
        """Set the number of components for this space.

        Logically collective.

        Parameters
        ----------
        nc
            The number of components.

        See Also
        --------
        getNumComponents, petsc.PetscSpaceSetNumComponents

        """
        cdef PetscInt cnc = asInt(nc)
        CHKERR(PetscSpaceSetNumComponents(self.space, cnc))

    def getType(self) -> str:
        """Return the type of the space object.

        Not collective.

        See Also
        --------
        setType, petsc.PetscSpaceGetType

        """
        cdef PetscSpaceType cval = NULL
        CHKERR(PetscSpaceGetType(self.space, &cval))
        return bytes2str(cval)

    def setType(self, space_type: Type | str) -> Self:
        """Build a particular type of space.

        Collective.

        Parameters
        ----------
        space_type
            The kind of space.

        See Also
        --------
        getType, petsc.PetscSpaceSetType

        """
        cdef PetscSpaceType cval = NULL
        space_type = str2bytes(space_type, &cval)
        CHKERR(PetscSpaceSetType(self.space, cval))
        return self

    def getSumConcatenate(self) -> bool:
        """Return the concatenate flag for this space.

        Not collective.

        A concatenated sum space will have the number of components equal to
        the sum of the number of components of all subspaces.
        A non-concatenated, or direct sum space will have the same number of
        components as its subspaces.

        See Also
        --------
        setSumConcatenate, petsc.PetscSpaceSumGetConcatenate

        """
        cdef PetscBool concatenate = PETSC_FALSE
        CHKERR(PetscSpaceSumGetConcatenate(self.space, &concatenate))
        return toBool(concatenate)

    def setSumConcatenate(self, concatenate: bool) -> None:
        """Set the concatenate flag for this space.

        Logically collective.

        A concatenated sum space will have the number of components equal to
        the sum of the number of components of all subspaces.
        A non-concatenated, or direct sum space will have the same number of
        components as its subspaces.

        Parameters
        ----------
        concatenate
            `True` if subspaces are concatenated components,
            `False` if direct summands.

        See Also
        --------
        getSumConcatenate, petsc.PetscSpaceSumSetConcatenate

        """
        cdef PetscBool cconcatenate = asBool(concatenate)
        CHKERR(PetscSpaceSumSetConcatenate(self.space, cconcatenate))

    def getSumNumSubspaces(self) -> int:
        """Return the number of spaces in the sum.

        Not collective.

        See Also
        --------
        setSumNumSubspaces, petsc.PetscSpaceSumGetNumSubspaces

        """
        cdef PetscInt numSumSpaces = 0
        CHKERR(PetscSpaceSumGetNumSubspaces(self.space, &numSumSpaces))
        return toInt(numSumSpaces)

    def getSumSubspace(self, s: int) -> Space:
        """Return a space in the sum.

        Not collective.

        Parameters
        ----------
        s
            The space number.

        See Also
        --------
        setSumSubspace, petsc.PetscSpaceSumGetSubspace

        """
        cdef Space subsp = Space()
        cdef PetscInt cs = asInt(s)
        CHKERR(PetscSpaceSumGetSubspace(self.space, cs, &subsp.space))
        return subsp

    def setSumSubspace(self, s: int, Space subsp) -> None:
        """Set a space in the sum.

        Logically collective.

        Parameters
        ----------
        s
            The space number.
        subsp
            The number of spaces.

        See Also
        --------
        getSumSubspace, petsc.PetscSpaceSumSetSubspace

        """
        cdef PetscInt cs = asInt(s)
        CHKERR(PetscSpaceSumSetSubspace(self.space, cs, subsp.space))

    def setSumNumSubspaces(self, numSumSpaces: int) -> None:
        """Set the number of spaces in the sum.

        Logically collective.

        Parameters
        ----------
        numSumSpaces
            The number of spaces.

        See Also
        --------
        getSumNumSubspaces, petsc.PetscSpaceSumSetNumSubspaces

        """
        cdef PetscInt cnumSumSpaces = asInt(numSumSpaces)
        CHKERR(PetscSpaceSumSetNumSubspaces(self.space, cnumSumSpaces))

    def getTensorNumSubspaces(self) -> int:
        """Return the number of spaces in the tensor product.

        Not collective.

        See Also
        --------
        setTensorNumSubspaces, petsc.PetscSpaceTensorGetNumSubspaces

        """
        cdef PetscInt cnumTensSpaces = 0
        CHKERR(PetscSpaceTensorGetNumSubspaces(self.space, &cnumTensSpaces))
        return toInt(cnumTensSpaces)

    def setTensorSubspace(self, s: int, Space subsp) -> None:
        """Set a space in the tensor product.

        Logically collective.

        Parameters
        ----------
        s
            The space number.
        subsp
            The number of spaces.

        See Also
        --------
        getTensorSubspace, petsc.PetscSpaceTensorSetSubspace

        """
        cdef PetscInt cs = asInt(s)
        CHKERR(PetscSpaceTensorSetSubspace(self.space, cs, subsp.space))

    def getTensorSubspace(self, s: int) -> Space:
        """Return a space in the tensor product.

        Not collective.

        Parameters
        ----------
        s
            The space number.

        See Also
        --------
        setTensorSubspace, petsc.PetscSpaceTensorGetSubspace

        """
        cdef PetscInt cs = asInt(s)
        cdef Space subsp = Space()
        CHKERR(PetscSpaceTensorGetSubspace(self.space, cs, &subsp.space))
        return subsp

    def setTensorNumSubspaces(self, numTensSpaces: int) -> None:
        """Set the number of spaces in the tensor product.

        Logically collective.

        Parameters
        ----------
        numTensSpaces
            The number of spaces.

        See Also
        --------
        getTensorNumSubspaces, petsc.PetscSpaceTensorSetNumSubspaces

        """
        cdef PetscInt cnumTensSpaces = asInt(numTensSpaces)
        CHKERR(PetscSpaceTensorSetNumSubspaces(self.space, cnumTensSpaces))

    def getPolynomialTensor(self) -> bool:
        """Return whether a function space is a space of tensor polynomials.

        Not collective.

        Return `True` if a function space is a space of tensor polynomials
        (the space is spanned by polynomials whose degree in each variable is
        bounded by the given order), as opposed to polynomials (the space is
        spanned by polynomials whose total degree—summing over all variables
        is bounded by the given order).

        See Also
        --------
        setPolynomialTensor, petsc.PetscSpacePolynomialGetTensor

        """
        cdef PetscBool ctensor = PETSC_FALSE
        CHKERR(PetscSpacePolynomialGetTensor(self.space, &ctensor))
        return toBool(ctensor)

    def setPolynomialTensor(self, tensor: bool) -> None:
        """Set whether a function space is a space of tensor polynomials.

        Logically collective.

        Set to `True` for a function space which is a space of tensor
        polynomials (the space is spanned by polynomials whose degree in each
        variable is bounded by the given order), as opposed to polynomials
        (the space is spanned by polynomials whose total degree—summing over
        all variables is bounded by the given order).

        Parameters
        ----------
        tensor
            `True` for a tensor polynomial space, `False` for a polynomial
            space.

        See Also
        --------
        getPolynomialTensor, petsc.PetscSpacePolynomialSetTensor

        """
        cdef PetscBool ctensor = asBool(tensor)
        CHKERR(PetscSpacePolynomialSetTensor(self.space, ctensor))

    def setPointPoints(self, Quad quad) -> None:
        """Set the evaluation points for the space to be based on a quad.

        Logically collective.

        Sets the evaluation points for the space to coincide with the points
        of a quadrature rule.

        Parameters
        ----------
        quad
            The `Quad` defining the points.

        See Also
        --------
        getPointPoints, petsc.PetscSpacePointSetPoints

        """
        CHKERR(PetscSpacePointSetPoints(self.space, quad.quad))

    def getPointPoints(self) -> Quad:
        """Return the evaluation points for the space as the points of a quad.

        Logically collective.

        See Also
        --------
        setPointPoints, petsc.PetscSpacePointGetPoints

        """
        cdef Quad quad = Quad()
        CHKERR(PetscSpacePointGetPoints(self.space, &quad.quad))
        CHKERR(PetscINCREF(quad.obj))
        return quad

    def setPTrimmedFormDegree(self, formDegree: int) -> None:
        """Set the form degree of the trimmed polynomials.

        Logically collective.

        Parameters
        ----------
        formDegree
            The form degree.

        See Also
        --------
        getPTrimmedFormDegree, petsc.PetscSpacePTrimmedSetFormDegree

        """
        cdef PetscInt cformDegree = asInt(formDegree)
        CHKERR(PetscSpacePTrimmedSetFormDegree(self.space, cformDegree))

    def getPTrimmedFormDegree(self) -> int:
        """Return the form degree of the trimmed polynomials.

        Not collective.

        See Also
        --------
        setPTrimmedFormDegree, petsc.PetscSpacePTrimmedGetFormDegree

        """
        cdef PetscInt cformDegree = 0
        CHKERR(PetscSpacePTrimmedGetFormDegree(self.space, &cformDegree))
        return toInt(cformDegree)

# --------------------------------------------------------------------


class DualSpaceType(object):
    """The dual space types."""
    LAGRANGE = S_(PETSCDUALSPACELAGRANGE)
    SIMPLE   = S_(PETSCDUALSPACESIMPLE)
    REFINED  = S_(PETSCDUALSPACEREFINED)
    BDM      = S_(PETSCDUALSPACEBDM)

# --------------------------------------------------------------------


cdef class DualSpace(Object):
    """Dual space to a linear space."""

    Type = DualSpaceType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.dualspace
        self.dualspace  = NULL

    def setUp(self) -> None:
        """Construct a basis for a `DualSpace`.

        Collective.

        See Also
        --------
        petsc.PetscDualSpaceSetUp

        """
        CHKERR(PetscDualSpaceSetUp(self.dualspace))

    def create(self, comm: Comm | None = None) -> Self:
        """Create an empty `DualSpace` object.

        Collective.

        The type can then be set with `setType`.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.PetscDualSpaceCreate

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscDualSpace newdsp = NULL
        CHKERR(PetscDualSpaceCreate(ccomm, &newdsp))
        CHKERR(PetscCLEAR(self.obj)); self.dualspace = newdsp
        return self

    def view(self, Viewer viewer=None) -> None:
        """View a `DualSpace`.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` to display the `DualSpace`.

        See Also
        --------
        petsc.PetscDualSpaceView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(PetscDualSpaceView(self.dualspace, vwr))

    def destroy(self) -> Self:
        """Destroy the `DualSpace` object.

        Collective.

        See Also
        --------
        petsc.PetscDualSpaceDestroy

        """
        CHKERR(PetscDualSpaceDestroy(&self.dualspace))
        return self

    def duplicate(self) -> DualSpace:
        """Create a duplicate `DualSpace` object that is not set up.

        Collective.

        See Also
        --------
        petsc.PetscDualSpaceDuplicate

        """
        cdef DualSpace spNew = DualSpace()
        CHKERR(PetscDualSpaceDuplicate(self.dualspace, &spNew.dualspace))

    def getDM(self) -> DM:
        """Return the `DM` representing the reference cell of a `DualSpace`.

        Not collective.

        See Also
        --------
        setDM, petsc.PetscDualSpaceGetDM

        """
        cdef PetscDM newdm = NULL
        CHKERR(PetscDualSpaceGetDM(self.dualspace, &newdm))
        cdef DM dm = subtype_DM(newdm)()
        dm.dm = newdm
        CHKERR(PetscINCREF(dm.obj))
        return dm

    def setDM(self, DM dm) -> None:
        """Set the `DM` representing the reference cell.

        Not collective.

        Parameters
        ----------
        dm
            The reference cell.

        See Also
        --------
        getDM, petsc.PetscDualSpaceSetDM

        """
        CHKERR(PetscDualSpaceSetDM(self.dualspace, dm.dm))

    def getDimension(self) -> int:
        """Return the dimension of the dual space.

        Not collective.

        The dimension of the dual space, i.e. the number of basis functionals.

        See Also
        --------
        petsc.PetscDualSpaceGetDimension

        """
        cdef PetscInt cdim = 0
        CHKERR(PetscDualSpaceGetDimension(self.dualspace, &cdim))
        return toInt(cdim)

    def getNumComponents(self) -> int:
        """Return the number of components for this space.

        Not collective.

        See Also
        --------
        setNumComponents, petsc.PetscDualSpaceGetNumComponents

        """
        cdef PetscInt cncomps = 0
        CHKERR(PetscDualSpaceGetNumComponents(self.dualspace, &cncomps))
        return toInt(cncomps)

    def setNumComponents(self, nc: int) -> None:
        """Set the number of components for this space.

        Logically collective.

        Parameters
        ----------
        nc
            The number of components

        See Also
        --------
        getNumComponents, petsc.PetscDualSpaceSetNumComponents

        """
        cdef PetscInt cnc = asInt(nc)
        CHKERR(PetscDualSpaceSetNumComponents(self.dualspace, cnc))

    def getType(self) -> str:
        """Return the type of the dual space object.

        Not collective.

        See Also
        --------
        setType, petsc.PetscDualSpaceGetType

        """
        cdef PetscDualSpaceType cval = NULL
        CHKERR(PetscDualSpaceGetType(self.dualspace, &cval))
        return bytes2str(cval)

    def setType(self, dualspace_type: Type | str) -> Self:
        """Build a particular type of dual space.

        Collective.

        Parameters
        ----------
        dualspace_type
            The kind of space.

        See Also
        --------
        getType, petsc.PetscDualSpaceSetType

        """
        cdef PetscDualSpaceType cval = NULL
        dualspace_type = str2bytes(dualspace_type, &cval)
        CHKERR(PetscDualSpaceSetType(self.dualspace, cval))
        return self

    def getOrder(self) -> int:
        """Return the order of the dual space.

        Not collective.

        See Also
        --------
        setOrder, petsc.PetscDualSpaceGetOrder

        """
        cdef PetscInt corder = 0
        CHKERR(PetscDualSpaceGetOrder(self.dualspace, &corder))
        return toInt(corder)

    def setOrder(self, order: int) -> None:
        """Set the order of the dual space.

        Not collective.

        Parameters
        ----------
        order
            The order.

        See Also
        --------
        getOrder, petsc.PetscDualSpaceSetOrder

        """
        cdef PetscInt corder = asInt(order)
        CHKERR(PetscDualSpaceSetOrder(self.dualspace, corder))

    def getNumDof(self) -> ArrayInt:
        """Return the number of degrees of freedom for each spatial dimension.

        Not collective.

        See Also
        --------
        petsc.PetscDualSpaceGetNumDof

        """
        cdef const PetscInt *cndof = NULL
        cdef PetscInt cdim = 0
        CHKERR(PetscDualSpaceGetDimension(self.dualspace, &cdim))
        CHKERR(PetscDualSpaceGetNumDof(self.dualspace, &cndof))
        return array_i(cdim + 1, cndof)

    def getFunctional(self, i: int) -> Quad:
        """Return the i-th basis functional in the dual space.

        Not collective.

        Parameters
        ----------
        i
            The basis number.

        See Also
        --------
        petsc.PetscDualSpaceGetFunctional

        """
        cdef PetscInt ci = asInt(i)
        cdef Quad functional = Quad()
        CHKERR(PetscDualSpaceGetFunctional(self.dualspace, ci, &functional.quad))
        CHKERR(PetscINCREF(functional.obj))
        return functional

    def getInteriorDimension(self) -> int:
        """Return the interior dimension of the dual space.

        Not collective.

        The interior dimension of the dual space, i.e. the number of basis
        functionals assigned to the interior of the reference domain.

        See Also
        --------
        petsc.PetscDualSpaceGetInteriorDimension

        """
        cdef PetscInt cintdim = 0
        CHKERR(PetscDualSpaceGetInteriorDimension(self.dualspace, &cintdim))
        return toInt(cintdim)

    def getLagrangeContinuity(self) -> bool:
        """Return whether the element is continuous.

        Not collective.

        See Also
        --------
        setLagrangeContinuity, petsc.PetscDualSpaceLagrangeGetContinuity

        """
        cdef PetscBool ccontinuous = PETSC_FALSE
        CHKERR(PetscDualSpaceLagrangeGetContinuity(self.dualspace, &ccontinuous))
        return toBool(ccontinuous)

    def setLagrangeContinuity(self, continuous: bool) -> None:
        """Indicate whether the element is continuous.

        Not collective.

        Parameters
        ----------
        continuous
            The flag for element continuity.

        See Also
        --------
        getLagrangeContinuity, petsc.PetscDualSpaceLagrangeSetContinuity

        """
        cdef PetscBool ccontinuous = asBool(continuous)
        CHKERR(PetscDualSpaceLagrangeSetContinuity(self.dualspace, ccontinuous))

    def getLagrangeTensor(self) -> bool:
        """Return the tensor nature of the dual space.

        Not collective.

        See Also
        --------
        setLagrangeTensor, petsc.PetscDualSpaceLagrangeGetTensor

        """
        cdef PetscBool ctensor = PETSC_FALSE
        CHKERR(PetscDualSpaceLagrangeGetTensor(self.dualspace, &ctensor))
        return toBool(ctensor)

    def setLagrangeTensor(self, tensor: bool) -> None:
        """Set the tensor nature of the dual space.

        Not collective.

        Parameters
        ----------
        tensor
            Whether the dual space has tensor layout (vs. simplicial).

        See Also
        --------
        getLagrangeTensor, petsc.PetscDualSpaceLagrangeSetTensor

        """
        cdef PetscBool ctensor = asBool(tensor)
        CHKERR(PetscDualSpaceLagrangeSetTensor(self.dualspace, ctensor))

    def getLagrangeTrimmed(self) -> bool:
        """Return the trimmed nature of the dual space.

        Not collective.

        See Also
        --------
        setLagrangeTrimmed, petsc.PetscDualSpaceLagrangeGetTrimmed

        """
        cdef PetscBool ctrimmed = PETSC_FALSE
        CHKERR(PetscDualSpaceLagrangeGetTrimmed(self.dualspace, &ctrimmed))
        return toBool(ctrimmed)

    def setLagrangeTrimmed(self, trimmed: bool) -> None:
        """Set the trimmed nature of the dual space.

        Not collective.

        Parameters
        ----------
        trimmed
            Whether the dual space represents to dual basis of a trimmed
            polynomial space (e.g. Raviart-Thomas and higher order /
            other form degree variants).

        See Also
        --------
        getLagrangeTrimmed, petsc.PetscDualSpaceLagrangeSetTrimmed

        """
        cdef PetscBool ctrimmed = asBool(trimmed)
        CHKERR(PetscDualSpaceLagrangeSetTrimmed(self.dualspace, ctrimmed))

    def setSimpleDimension(self, dim: int) -> None:
        """Set the number of functionals in the dual space basis.

        Logically collective.

        Parameters
        ----------
        dim
            The basis dimension.

        See Also
        --------
        petsc.PetscDualSpaceSimpleSetDimension

        """
        cdef PetscInt cdim = asInt(dim)
        CHKERR(PetscDualSpaceSimpleSetDimension(self.dualspace, cdim))

    def setSimpleFunctional(self, func: int, Quad functional) -> None:
        """Set the given basis element for this dual space.

        Not collective.

        Parameters
        ----------
        func
            The basis index.
        functional
            The basis functional.

        See Also
        --------
        petsc.PetscDualSpaceSimpleSetFunctional

        """
        cdef PetscInt cfunc = asInt(func)
        CHKERR(PetscDualSpaceSimpleSetFunctional(self.dualspace, cfunc, functional.quad))

del SpaceType
del DualSpaceType



# --------------------------------------------------------------------

cdef extern from * nogil:
    """
    #include "pyapicompat.h"
    #include "lib-petsc/compat.h"
    #include "lib-petsc/custom.h"

    /* Silence Clang warnings in Cython-generated C code */
    #if defined(__clang__)
      #pragma clang diagnostic ignored "-Wextra-semi-stmt"
      #pragma clang diagnostic ignored "-Wparentheses-equality"
      #pragma clang diagnostic ignored "-Wunreachable-code-fallthrough"
      #pragma clang diagnostic ignored "-Woverlength-strings"
      #pragma clang diagnostic ignored "-Wunreachable-code"
      #pragma clang diagnostic ignored "-Wundef"
    #elif defined(__GNUC__) || defined(__GNUG__)
      #pragma GCC diagnostic ignored "-Wstrict-aliasing"
      #pragma GCC diagnostic ignored "-Wtype-limits"
    #endif
    """

# --------------------------------------------------------------------

cdef extern from * nogil:
    ctypedef ssize_t Py_intptr_t
    ctypedef size_t  Py_uintptr_t

# --------------------------------------------------------------------

cdef inline object bytes2str(const char p[]):
    if p == NULL:
        return None
    cdef bytes s = <char*>p
    if isinstance(s, str):
        return s
    else:
        return s.decode()

cdef inline object str2bytes(object s, const char *p[]):
    if s is None:
        p[0] = NULL
        return None
    if not isinstance(s, bytes):
        s = s.encode()
    p[0] = <const char*>(<char*>s)
    return s

cdef inline object S_(const char p[]):
    if p == NULL: return None
    cdef object s = <char*>p
    return s if isinstance(s, str) else s.decode()


# --------------------------------------------------------------------

# SETERR Support
# --------------

cdef extern from "Python.h":
    void PyErr_SetObject(object, object)
    PyObject *PyExc_RuntimeError
    PyObject *PyErr_GetRaisedException()
    void PyErr_SetRaisedException(PyObject*)
    void PyException_SetCause(PyObject*, PyObject*)

cdef object PetscError = <object>PyExc_RuntimeError

cdef inline void PetscTracebackAdd(object exc) noexcept:
    if (<void*>PetscError) == NULL: return
    global tracebacklist
    exc._traceback[:] = tracebacklist[:]
    del tracebacklist[:]

cdef inline int SETERR(PetscErrorCode ierr) noexcept nogil:
    cdef PyObject *exception = NULL, *cause = NULL
    with gil:
        cause = PyErr_GetRaisedException()
        if (<void*>PetscError) != NULL:
            PyErr_SetObject(PetscError, <long>ierr)
        else:
            PyErr_SetObject(<object>PyExc_RuntimeError, <long>ierr)
        exception = PyErr_GetRaisedException()
        PetscTracebackAdd(<object>exception)
        PyException_SetCause(exception, cause)
        PyErr_SetRaisedException(exception)
    return 0

cdef inline PetscErrorCode CHKERR(PetscErrorCode ierr) except PETSC_ERR_PYTHON nogil:
    if ierr == PETSC_SUCCESS:
        return PETSC_SUCCESS # no error
    <void>SETERR(ierr)
    return PETSC_ERR_PYTHON

# SETERRMPI Support
# -----------------

cdef extern from * nogil:
    enum: MPI_SUCCESS
    enum: MPI_MAX_ERROR_STRING
    int MPI_Error_string(int, char[], int*)
    PetscErrorCode PetscSNPrintf(char[], size_t, const char[], ...)
    PetscErrorCode PetscERROR(MPI_Comm, const char[], PetscErrorCode, int, const char[], const char[])

cdef inline int SETERRMPI(int ierr) noexcept nogil:
    cdef char mpi_err_str[MPI_MAX_ERROR_STRING]
    cdef int  result_len = <int>sizeof(mpi_err_str)
    <void>memset(mpi_err_str, 0, <size_t>result_len)
    <void>MPI_Error_string(ierr, mpi_err_str, &result_len)
    <void>result_len  # unused-but-set-variable
    cdef char error_str[MPI_MAX_ERROR_STRING+64]
    <void>PetscSNPrintf(error_str, sizeof(error_str), b"MPI Error %s %d", mpi_err_str, ierr)
    <void>PetscERROR(PETSC_COMM_SELF, "Unknown Python Function", PETSC_ERR_MPI, PETSC_ERROR_INITIAL, "%s", error_str)
    <void>SETERR(PETSC_ERR_MPI)
    return 0

cdef inline PetscErrorCode CHKERRMPI(int ierr) except PETSC_ERR_PYTHON nogil:
    if ierr == MPI_SUCCESS:
        return PETSC_SUCCESS
    <void>SETERRMPI(ierr)
    return PETSC_ERR_PYTHON

# --------------------------------------------------------------------

# PETSc support
# -------------

cdef extern from * nogil:
    ctypedef bint      PetscBool
    const    PetscBool PETSC_TRUE
    const    PetscBool PETSC_FALSE
    ctypedef long      PetscInt
    ctypedef long long PetscInt64
    ctypedef double    PetscReal
    ctypedef double    PetscScalar
    ctypedef ptrdiff_t PetscCount

cdef extern from "<petsc4py/pyscalar.h>":
    object      PyPetscScalar_FromPetscScalar(PetscScalar)
    PetscScalar PyPetscScalar_AsPetscScalar(object) except? <PetscScalar>-1.0

cdef extern from "<petsc4py/pybuffer.h>":
    int  PyPetscBuffer_FillInfo(Py_buffer*, void*, PetscInt, char, int, int) except -1
    void PyPetscBuffer_Release(Py_buffer*)

cdef inline object toBool(PetscBool value):
    return True if value == PETSC_TRUE else False
cdef inline PetscBool asBool(object value) except? <PetscBool>0:
    return PETSC_TRUE if value else PETSC_FALSE

cdef inline object toInt(PetscInt value):
    return value
cdef inline PetscInt asInt(object value) except? -1:
    return value

cdef inline object toReal(PetscReal value):
    return value
cdef inline PetscReal asReal(object value) except? -1:
    return value

cdef inline object toScalar(PetscScalar value):
    return PyPetscScalar_FromPetscScalar(value)
cdef inline PetscScalar asScalar(object value) except? <PetscScalar>-1.0:
    return PyPetscScalar_AsPetscScalar(value)

# --------------------------------------------------------------------

# NumPy support
# -------------

include "arraynpy.pxi"

import_array()

IntType     = PyArray_TypeObjectFromType(NPY_PETSC_INT)
RealType    = PyArray_TypeObjectFromType(NPY_PETSC_REAL)
ScalarType  = PyArray_TypeObjectFromType(NPY_PETSC_SCALAR)
ComplexType = PyArray_TypeObjectFromType(NPY_PETSC_COMPLEX)

include "dlpack.pxi"

# --------------------------------------------------------------------

include "typing.pxi"
include "petscdef.pxi"
include "petscmem.pxi"
include "petscopt.pxi"
include "petscmpi.pxi"
include "petscsys.pxi"
include "petsclog.pxi"
include "petscobj.pxi"
include "petscvwr.pxi"
include "petscrand.pxi"
include "petscdevice.pxi"
include "petsclayout.pxi"
include "petscis.pxi"
include "petscsf.pxi"
include "petscvec.pxi"
include "petscdt.pxi"
include "petscfe.pxi"
include "petscsct.pxi"
include "petscsec.pxi"
include "petscmat.pxi"
include "petscmatpartitioning.pxi"
include "petscpc.pxi"
include "petscksp.pxi"
include "petscsnes.pxi"
include "petscts.pxi"
include "petsctao.pxi"
include "petscao.pxi"
include "petscdm.pxi"
include "petscds.pxi"
include "petscdmda.pxi"
include "petscdmplex.pxi"
include "petscdmstag.pxi"
include "petscdmcomposite.pxi"
include "petscdmshell.pxi"
include "petscdmlabel.pxi"
include "petscdmswarm.pxi"
include "petscpartitioner.pxi"
include "petscspace.pxi"
include "petscdmutils.pxi"
include "petscpyappctx.pxi"
include "petscregressor.pxi"

# --------------------------------------------------------------------

__doc__ = """
Portable, Extensible Toolkit for Scientific Computation.
"""

include "Const.pyx"
include "Error.pyx"
include "Options.pyx"
include "Sys.pyx"
include "Log.pyx"
include "Comm.pyx"
include "Object.pyx"
include "Viewer.pyx"
include "Random.pyx"
include "Device.pyx"
include "IS.pyx"
include "SF.pyx"
include "Vec.pyx"
include "DT.pyx"
include "FE.pyx"
include "Scatter.pyx"
include "Section.pyx"
include "Mat.pyx"
include "MatPartitioning.pyx"
include "PC.pyx"
include "KSP.pyx"
include "SNES.pyx"
include "TS.pyx"
include "TAO.pyx"
include "AO.pyx"
include "DM.pyx"
include "DS.pyx"
include "DMDA.pyx"
include "DMPlex.pyx"
include "DMStag.pyx"
include "DMComposite.pyx"
include "DMShell.pyx"
include "DMLabel.pyx"
include "DMSwarm.pyx"
include "Partitioner.pyx"
include "Space.pyx"
include "DMUtils.pyx"
include "Regressor.pyx"

# --------------------------------------------------------------------

include "CAPI.pyx"
include "libpetsc4py.pyx"

# --------------------------------------------------------------------

cdef extern from "Python.h":
    int Py_IsInitialized() nogil
    int PyList_Insert(object, Py_ssize_t, object) except -1
    int PyList_Append(object, object) except -1

cdef extern from * nogil:
    PetscErrorCode PetscTBEH(MPI_Comm, int, char*, char*, int, PetscErrorType, char*, void*)

cdef object tracebacklist = []

cdef PetscErrorCode traceback(
    MPI_Comm       comm,
    int            line,
    const char    *cfunc,
    const char    *cfile,
    PetscErrorCode n,
    PetscErrorType p,
    const char    *mess,
    void          *ctx,
) except (<PetscErrorCode>-1) with gil:
    cdef PetscLogDouble mem=0
    cdef PetscLogDouble rss=0
    cdef const char    *text=NULL
    global tracebacklist
    cdef object tbl = tracebacklist
    cdef object fun = bytes2str(cfunc)
    cdef object fnm = bytes2str(cfile)
    cdef object m = "%s() at %s:%d" % (fun, fnm, line)
    PyList_Insert(tbl, 0, m)
    if p != PETSC_ERROR_INITIAL:
        return n
    #
    del tbl[1:] # clear any previous stuff
    if n == PETSC_ERR_MEM: # special case
        PetscMallocGetCurrentUsage(&mem)
        PetscMemoryGetCurrentUsage(&rss)
        m = (
            "Out of memory. "
            "Allocated: %d, "
            "Used by process: %d"
        ) % (mem, rss)
        PyList_Append(tbl, m)
    else:
        PetscErrorMessage(n, &text, NULL)
    if text != NULL: PyList_Append(tbl, bytes2str(text))
    if mess != NULL: PyList_Append(tbl, bytes2str(mess))
    <void>comm # unused
    <void>ctx  # unused
    return n

cdef PetscErrorCode PetscPythonErrorHandler(
    MPI_Comm       comm,
    int            line,
    const char    *cfunc,
    const char    *cfile,
    PetscErrorCode n,
    PetscErrorType p,
    const char    *mess,
    void          *ctx,
) except (<PetscErrorCode>-1) nogil:
    global tracebacklist
    if (<void*>tracebacklist) != NULL and Py_IsInitialized():
        return traceback(comm, line, cfunc, cfile, n, p, mess, ctx)
    else:
        return PetscTBEH(comm, line, cfunc, cfile, n, p, mess, ctx)

# --------------------------------------------------------------------

cdef extern from "<stdlib.h>" nogil:
    void* malloc(size_t)
    void* realloc (void*, size_t)
    void free(void*)

cdef extern from "<stdarg.h>" nogil:
    ctypedef struct va_list:
        pass

cdef extern from "<string.h>" nogil:
    void* memset(void*, int, size_t)
    void* memcpy(void*, void*, size_t)
    char* strdup(char*)

cdef extern from "<stdio.h>" nogil:
    ctypedef struct FILE
    FILE *stderr
    int fprintf(FILE *, char *, ...)

cdef extern from "Python.h":
    int Py_AtExit(void (*)() noexcept nogil)
    void PySys_WriteStderr(char*, ...)

cdef extern from * nogil:
    """
    #include "lib-petsc/initpkg.h"
    """
    PetscErrorCode PetscInitializePackageAll()

cdef int    PyPetsc_Argc = 0
cdef char** PyPetsc_Argv = NULL

cdef int getinitargs(object args, int *argc, char **argv[]) except -1:
    # allocate command line arguments
    cdef int i, c = 0
    cdef char **v = NULL
    if args is None: args = []
    args = [str(a).encode() for a in args]
    args = [a for a in args if a]
    c = <int> len(args)
    v = <char**> malloc(<size_t>(c+1)*sizeof(char*))
    if v == NULL: raise MemoryError
    memset(v, 0, <size_t>(c+1)*sizeof(char*))
    try:
        for 0 <= i < c:
            v[i] = strdup(args[i])
            if v[i] == NULL:
                raise MemoryError
    except Exception:
        delinitargs(&c, &v); raise
    argc[0] = c; argv[0] = v
    return 0

cdef void delinitargs(int *argc, char **argv[]) noexcept nogil:
    # deallocate command line arguments
    cdef int i, c = argc[0]
    cdef char** v = argv[0]
    argc[0] = 0; argv[0] = NULL
    if c >= 0 and v != NULL:
        for 0 <= i < c:
            if  v[i] != NULL: free(v[i])
        free(v)

cdef void finalize() noexcept nogil:
    cdef int ierr = 0
    # deallocate command line arguments
    global PyPetsc_Argc, PyPetsc_Argv
    global PetscVFPrintf, prevfprintf
    delinitargs(&PyPetsc_Argc, &PyPetsc_Argv)
    # manage PETSc finalization
    if not (<int>PetscInitializeCalled): return
    if (<int>PetscFinalizeCalled): return
    # stop stdout/stderr redirect
    if (prevfprintf != NULL):
        PetscVFPrintf = prevfprintf
        prevfprintf = NULL
    # deinstall Python error handler
    ierr = PetscPopErrorHandler()
    if ierr != 0:
        fprintf(stderr,
                "PetscPopErrorHandler() failed "
                "[error code: %d]\n", ierr)
    # finalize PETSc
    ierr = PetscFinalize()
    if ierr != 0:
        fprintf(stderr,
                "PetscFinalize() failed "
                "[error code: %d]\n", ierr)
    # and we are done, see you later !!

# --------------------------------------------------------------------

cdef extern from *:
    PetscErrorCode (*PetscVFPrintf)(FILE*, const char*, va_list) except PETSC_ERR_PYTHON nogil

cdef PetscErrorCode (*prevfprintf)(FILE*, const char*, va_list) except PETSC_ERR_PYTHON nogil
prevfprintf = NULL

cdef PetscErrorCode PetscVFPrintf_PythonStdStream(
    FILE *fd, const char fmt[], va_list ap,
) except PETSC_ERR_PYTHON with gil:
    import sys
    cdef char cstring[8192]
    cdef size_t stringlen = sizeof(cstring)
    cdef size_t final_pos = 0
    if (fd == PETSC_STDOUT) and not (sys.stdout == sys.__stdout__):
        CHKERR(PetscVSNPrintf(&cstring[0], stringlen, fmt, &final_pos, ap))
        if final_pos > 0 and cstring[final_pos-1] == '\x00':
            final_pos -= 1
        ustring = cstring[:final_pos].decode('UTF-8')
        sys.stdout.write(ustring)
    elif (fd == PETSC_STDERR) and not (sys.stderr == sys.__stderr__):
        CHKERR(PetscVSNPrintf(&cstring[0], stringlen, fmt, &final_pos, ap))
        if final_pos > 0 and cstring[final_pos-1] == '\x00':
            final_pos -= 1
        ustring = cstring[:final_pos].decode('UTF-8')
        sys.stderr.write(ustring)
    else:
        CHKERR(PetscVFPrintfDefault(fd, fmt, ap))
    return PETSC_SUCCESS

cdef int _push_vfprintf(
    PetscErrorCode (*vfprintf)(FILE*, const char*, va_list) except PETSC_ERR_PYTHON nogil,
) except -1:
    global PetscVFPrintf, prevfprintf
    assert prevfprintf == NULL
    prevfprintf = PetscVFPrintf
    PetscVFPrintf = vfprintf

cdef int _pop_vfprintf() except -1:
    global PetscVFPrintf, prevfprintf
    assert prevfprintf != NULL
    PetscVFPrintf = prevfprintf
    prevfprintf == NULL

cdef int initialize(object args, object comm) except -1:
    if (<int>PetscInitializeCalled): return 1
    if (<int>PetscFinalizeCalled):   return 0
    # allocate command line arguments
    global PyPetsc_Argc, PyPetsc_Argv
    getinitargs(args, &PyPetsc_Argc, &PyPetsc_Argv)
    # communicator
    global PETSC_COMM_WORLD
    PETSC_COMM_WORLD = def_Comm(comm, PETSC_COMM_WORLD)
    # initialize PETSc
    CHKERR(PetscInitialize(&PyPetsc_Argc, &PyPetsc_Argv, NULL, NULL))
    # install Python error handler
    cdef PetscErrorHandlerFunction handler = NULL
    handler = <PetscErrorHandlerFunction>PetscPythonErrorHandler
    CHKERR(PetscPushErrorHandler(handler, NULL))
    # redirect PETSc std streams
    import sys
    if (sys.stdout != sys.__stdout__) or (sys.stderr != sys.__stderr__):
        _push_vfprintf(&PetscVFPrintf_PythonStdStream)
    # register finalization function
    if Py_AtExit(finalize) < 0:
        PySys_WriteStderr(b"warning: could not register %s with Py_AtExit()",
                          b"PetscFinalize()")
    return 1 # and we are done, enjoy !!

cdef extern from * nogil:
    PetscClassId PETSC_OBJECT_CLASSID           "PETSC_OBJECT_CLASSID"
    PetscClassId PETSC_VIEWER_CLASSID           "PETSC_VIEWER_CLASSID"
    PetscClassId PETSC_RANDOM_CLASSID           "PETSC_RANDOM_CLASSID"
    PetscClassId PETSC_IS_CLASSID               "IS_CLASSID"
    PetscClassId PETSC_LGMAP_CLASSID            "IS_LTOGM_CLASSID"
    PetscClassId PETSC_SF_CLASSID               "PETSCSF_CLASSID"
    PetscClassId PETSC_VEC_CLASSID              "VEC_CLASSID"
    PetscClassId PETSC_SECTION_CLASSID          "PETSC_SECTION_CLASSID"
    PetscClassId PETSC_MAT_CLASSID              "MAT_CLASSID"
    PetscClassId PETSC_MAT_PARTITIONING_CLASSID "MAT_PARTITIONING_CLASSID"
    PetscClassId PETSC_NULLSPACE_CLASSID        "MAT_NULLSPACE_CLASSID"
    PetscClassId PETSC_PC_CLASSID               "PC_CLASSID"
    PetscClassId PETSC_KSP_CLASSID              "KSP_CLASSID"
    PetscClassId PETSC_SNES_CLASSID             "SNES_CLASSID"
    PetscClassId PETSC_TS_CLASSID               "TS_CLASSID"
    PetscClassId PETSC_TAO_CLASSID              "TAO_CLASSID"
    PetscClassId PETSC_AO_CLASSID               "AO_CLASSID"
    PetscClassId PETSC_DM_CLASSID               "DM_CLASSID"
    PetscClassId PETSC_DS_CLASSID               "PETSCDS_CLASSID"
    PetscClassId PETSC_PARTITIONER_CLASSID      "PETSCPARTITIONER_CLASSID"
    PetscClassId PETSC_FE_CLASSID               "PETSCFE_CLASSID"
    PetscClassId PETSC_DMLABEL_CLASSID          "DMLABEL_CLASSID"
    PetscClassId PETSC_SPACE_CLASSID            "PETSCSPACE_CLASSID"
    PetscClassId PETSC_DUALSPACE_CLASSID        "PETSCDUALSPACE_CLASSID"
    PetscClassId PETSC_DEVICE_CLASSID           "PETSC_DEVICE_CLASSID"
    PetscClassId PETSC_DEVICE_CONTEXT_CLASSID   "PETSC_DEVICE_CONTEXT_CLASSID"
    PetscClassId PETSC_REGRESSOR_CLASSID        "PETSCREGRESSOR_CLASSID"

cdef bint registercalled = 0

cdef const char *citation = b"""\
@Article{Dalcin2011,
  Author = {Lisandro D. Dalcin and Rodrigo R. Paz and Pablo A. Kler and Alejandro Cosimo},
  Title = {Parallel distributed computing using {P}ython},
  Journal = {Advances in Water Resources},
  Note = {New Computational Methods and Software Tools},
  Volume = {34},
  Number = {9},
  Pages = {1124--1139},
  Year = {2011},
  DOI = {https://doi.org/10.1016/j.advwatres.2011.04.013}
}
"""

cdef int register() except -1:
    global registercalled
    if registercalled: return 0
    registercalled = True
    # register citation
    CHKERR(PetscCitationsRegister(citation, NULL))
    # make sure all PETSc packages are initialized
    CHKERR(PetscInitializePackageAll())
    # register custom implementations
    CHKERR(PetscPythonRegisterAll())
    # register Python types
    PyPetscType_Register(PETSC_OBJECT_CLASSID,           Object)
    PyPetscType_Register(PETSC_VIEWER_CLASSID,           Viewer)
    PyPetscType_Register(PETSC_RANDOM_CLASSID,           Random)
    PyPetscType_Register(PETSC_DEVICE_CLASSID,           Device)
    PyPetscType_Register(PETSC_DEVICE_CONTEXT_CLASSID,   DeviceContext)
    PyPetscType_Register(PETSC_IS_CLASSID,               IS)
    PyPetscType_Register(PETSC_LGMAP_CLASSID,            LGMap)
    PyPetscType_Register(PETSC_SF_CLASSID,               SF)
    PyPetscType_Register(PETSC_VEC_CLASSID,              Vec)
    PyPetscType_Register(PETSC_SECTION_CLASSID,          Section)
    PyPetscType_Register(PETSC_MAT_CLASSID,              Mat)
    PyPetscType_Register(PETSC_MAT_PARTITIONING_CLASSID, MatPartitioning)
    PyPetscType_Register(PETSC_NULLSPACE_CLASSID,        NullSpace)
    PyPetscType_Register(PETSC_PC_CLASSID,               PC)
    PyPetscType_Register(PETSC_KSP_CLASSID,              KSP)
    PyPetscType_Register(PETSC_SNES_CLASSID,             SNES)
    PyPetscType_Register(PETSC_TS_CLASSID,               TS)
    PyPetscType_Register(PETSC_TAO_CLASSID,              TAO)
    PyPetscType_Register(PETSC_PARTITIONER_CLASSID,      Partitioner)
    PyPetscType_Register(PETSC_AO_CLASSID,               AO)
    PyPetscType_Register(PETSC_DM_CLASSID,               DM)
    PyPetscType_Register(PETSC_DS_CLASSID,               DS)
    PyPetscType_Register(PETSC_FE_CLASSID,               FE)
    PyPetscType_Register(PETSC_DMLABEL_CLASSID,          DMLabel)
    PyPetscType_Register(PETSC_SPACE_CLASSID,            Space)
    PyPetscType_Register(PETSC_DUALSPACE_CLASSID,        DualSpace)
    PyPetscType_Register(PETSC_REGRESSOR_CLASSID,        Regressor)
    return 0 # and we are done, enjoy !!

# --------------------------------------------------------------------


def _initialize(args=None, comm=None):
    import atexit
    global PetscError
    PetscError = Error
    #
    cdef int ready = initialize(args, comm)
    if ready: register()
    #
    global __COMM_SELF__, __COMM_WORLD__
    __COMM_SELF__.comm  = PETSC_COMM_SELF
    __COMM_WORLD__.comm = PETSC_COMM_WORLD
    #
    global PETSC_COMM_DEFAULT
    PETSC_COMM_DEFAULT = PETSC_COMM_WORLD
    # Register finalizer
    atexit.register(_pre_finalize)


def _pre_finalize():
    # Called while the Python interpreter is still running
    garbage_cleanup()


def _finalize():
    finalize()
    #
    global __COMM_SELF__
    __COMM_SELF__.comm  = MPI_COMM_NULL
    global __COMM_WORLD__
    __COMM_WORLD__.comm = MPI_COMM_NULL
    #
    global PETSC_COMM_DEFAULT
    PETSC_COMM_DEFAULT = MPI_COMM_NULL
    #
    global type_registry
    type_registry.clear()
    global stage_registry
    stage_registry.clear()
    global class_registry
    class_registry.clear()
    global event_registry
    event_registry.clear()
    global citations_registry
    citations_registry.clear()


def _push_python_vfprintf():
    _push_vfprintf(&PetscVFPrintf_PythonStdStream)


def _pop_python_vfprintf():
    _pop_vfprintf()


def _stdout_is_stderr():
    global PETSC_STDOUT, PETSC_STDERR
    return PETSC_STDOUT == PETSC_STDERR

# --------------------------------------------------------------------
